"""
Copyright (C) 2022 Corsair M360, Inc - All Rights Reserved.
Unauthorized copying of this file, via any medium is strictly prohibited.
Proprietary and confidential.
"""

import os
import copy
import json
import socket
import logging

from m360 import constants, utils

_logger = logging.getLogger(__name__)

class Service(object):
    """Offers service related helper methods"""

    # singleton
    _instance = None

    def __init__(self):
        _logger.debug("Service created")

    @staticmethod
    def service_singleton():
        """Access the singleton receiver"""

        if Service._instance:
            return Service._instance

        if not Service._instance:
            _logger.debug("Creating service instance")
            Service._instance = Service()

        return Service._instance

    @staticmethod
    def __load_contract(path_or_obj):
        """
        Load the contract dict either from a path with a JSON file or just clone the passed dict
        @param path_or_obj  {Union[dict, str]}  Either a contract dict or a path to a JSON file containing the data
        @returns {dict} The contract object
        """
        if not path_or_obj or (type(path_or_obj) is str and not path_or_obj.strip()):
            _logger.error("Missing Service Contract!")
            raise Exception("Missing Service Contract!")

        if type(path_or_obj) is str:
            contract = utils.json_file_to_dict(path_or_obj)
            if not contract or (type(contract) is dict and len(contract.keys()) == 0):
                _logger.error("Invalid service contract found in the provided file!")
                raise Exception("Invalid service contract found in the provided file!")
        else:
            try:
                contract = copy.deepcopy(path_or_obj)
            except Exception as e:
                _logger.error(e)
                raise Exception("Unable to parse the service contract that you provided as a JSON object!")
        return contract

    @staticmethod
    def __make_request(*, method="get", path, headers=None, body=None, limit):
        """
        Utility method to make calls to APIs for this specific class
        """
        # build url
        url = constants.GATEWAY_MAINTENANCE_IP
        if (":" + str(constants.GATEWAY_MAINTENANCE_PORT)) not in url:
            url += ":" + str(constants.GATEWAY_MAINTENANCE_PORT)
        url += path
        if "https://" not in url:
            url = "https://" + url

        # make request with retry strategy
        delay = 0.001 if constants.MW_TESTING else 5
        res = utils.retry_request(method=method, url=url, is_json=True, retries=(limit - 1), retry_delay=delay,
                                  headers=headers, body=body).json()
        return res.get("data")

    @staticmethod
    def get(path_or_obj):
        """
        Fetches the JSON copy that represents the service as it's registered in the API gateway
        @param path_or_obj  {Union[dict, str]}  Either a contract dict or a path to a JSON file containing the data
        @returns {dict} API call response
        """
        # load contract data
        contract = Service.__load_contract(path_or_obj)
        # call API
        return Service.__make_request(path="/service", limit=3, headers={"service": contract.get("name"),
                                                                         "version": str(contract.get("version"))})

    @staticmethod
    def register(path_or_obj, ip, config):
        """
        Formulates a payload from provided parameters and makes a POST request call to the gateway maintenance API
        to register the service.
        @param path_or_obj  {Union[dict, str]}  Either a contract dict or a path to a JSON file containing the data
        @param ip           {str}
        @param config       {dict}
        @returns {dict} API call response
        """
        # load contract data
        contract = Service.__load_contract(path_or_obj)
        # check IP
        if not ip or (type(ip) is str and not ip.strip()):
            _logger.error("Missing IP Address Value!")
            raise Exception("Missing IP Address Value!")

        if config and (type(config) is dict and config.get("platform") == "kubernetes"):
            if not config.get("platformOptions"):
                raise Exception("Missing 'platformOptions' in config")
            if not contract.get("ports"):
                contract["ports"] = {}
            exposed_port = config.get("platformOptions").get("exposedPort")
            contract["ports"]["data"] = exposed_port
            contract["ports"]["maintenance"] = exposed_port

        contract["host"] = {
            "weight": int(contract.get("host").get("weight")) if contract.get("host") and contract.get("host").get(
                "weight") else 1,
            "hostname": socket.gethostname(),
            "ip": ip
        }

        # call API
        return Service.__make_request(method="post", path="/service/register", limit=3, body=contract)

    @staticmethod
    def ensure_ip_address(config):
        """
        Ensures that the provided IP address value in the configuration object is based on the platform being used
        @param config   {dict}
        @returns {bool} API call response
        """
        if not config or type(config) is not dict:
            _logger.error("Invalid configuration")
            raise Exception("Invalid configuration")

        # if the platform hosting the microservice is kubernetes, the IP address is provided as an env variable
        # we do not ping the ip address because it will not be reachable from the pod
        if config.get("platform") == "kubernetes":
            return True

        # otherwise, ping the ip address and get the numeric_host from the response.
        # This works for manual deployment and other deployment.
        # This also works for docker deployment
        #      IF CONFIGURATION.IP EQUALS THE NAME OF THE DOCKER SWARM SERVICE
        # TODO: make sure the below works and is equivalent to the description above
        domain = config.get("ip").replace("http:", "").replace("https:", "").replace("/", "")
        _logger.debug("Getting IP address for domain: " + domain)
        ip = socket.gethostbyname(domain)
        _logger.debug("Domain's IP address is: " + ip)
        if ip:
            config["ip"] = ip

        return True

    @staticmethod
    def get_endpoint_info(request):
        """
        This method returns the request endpoint information that was detected by the gateway and injected in the
        headers
        @param request   {dict}
        @returns {dict}
        """
        if not request:
            return None
        headers = utils.get_req_headers(request)
        if not headers or not headers.get("m360"):
            return None
        try:
            m360 = headers.get("m360")
            if m360:
                m360 = utils.from_json(m360)
                return m360.get("API")
            return None
        except Exception as e:
            _logger.error(e)
            return None


def instance():
    """Get the singleton instance of the service"""
    _logger.debug("calling service_instance, os.pid=%d", os.getpid())
    return Service.service_singleton()



"""
Copyright (C) 2022 Corsair M360, Inc - All Rights Reserved.
Unauthorized copying of this file, via any medium is strictly prohibited.
Proprietary and confidential.
"""
import traceback
import logging
import docker
import subprocess
import re

from m360 import constants, utils

_logger = logging.getLogger(__name__)

class Maintenance(object):
    """Handles communication with the gateway"""

    _service_name = None
    _service_version = None

    def __init__(self, configuration, app_name, registry, service, awareness):

        self._configuration = configuration
        self._app_name = app_name
        self._registry = registry
        self._service = service
        self._awareness = awareness
        _logger.debug("Maintenance init complete")

    def reload_registry(self, ignore_list=None):
        try:
            res = self._registry.reload()
            self.notify_brothers("/registry/reload", ignore_list)
            return {
                "name": self._app_name,
                "registry": res
            }
        except Exception as e:
            _logger.error(traceback.format_exc())
            raise e

    def reload_awareness(self, ignore_list=None):
        try:
            res = self._awareness.reload(constants.GATEWAY_MAINTENANCE_IP, constants.GATEWAY_MAINTENANCE_PORT, {})
            self.notify_brothers("/awareness/reload", ignore_list)
            return {
                "name": self._app_name,
                "awareness": res
            }
        except Exception as e:
            _logger.error(e)
            raise e

    @staticmethod
    def get_current_ip():
        my_ip = None
        try:
            result = subprocess.run(['hostname', '-i'], stdout=subprocess.PIPE)
            my_ip = re.sub(r"[\n\t\s]*", "", result.stdout.decode('utf-8'))
        except Exception as e:
            _logger.error(e)
        _logger.debug("Found current host ip: " + my_ip)
        return my_ip

    def get_brothers(self, ignore_list):
        """Invokes gateway APIs"""
        ignore_list = ignore_list.copy() if ignore_list else []
        res = {
            "brothers": [],
            "ignoreList": ignore_list
        }
        platform = self._configuration.get("platform")
        # skip logic if not in docker or kube env
        if platform != "docker" and platform != "kubernetes":
            _logger.debug("Skipping get_brothers because platform is: " + platform)
            return res
        pf_options = self._configuration.get("platformOptions")
        if not pf_options or not isinstance(pf_options, dict):
            _logger.error("Missing 'platformOptions' from config: " + str(self._configuration))
            raise Exception("Configuration must specify 'platformOptions' when platform is '" + platform + "'")

        my_ip = self.get_current_ip()
        if my_ip:
            ignore_list.append(my_ip)

        if platform == "kubernetes":
            # ".subsets[].addresses[].ip"
            #
            # TOKEN=$(cat /var/run/secrets/kubernetes.io/serviceaccount/token); \
            # curl https://kubernetes.default.svc/api/v1/namespaces/${namespace}/endpoints/${service-name} \
            #    --header "Authorization: Bearer $TOKEN" --insecure

            # read token from file
            with open(constants.SVC_ACCOUNT_TOKEN_PATH, mode="r", encoding="utf8") as f:
                kube_token = f.read()
            # get sibling IPs
            api_url = "https://kubernetes.default.svc/api/v1/namespaces/" + \
                      self._configuration.get("platformOptions").get("namespace") + "/endpoints/" + \
                      self._configuration.get("platformOptions").get("service")
            api_res = utils.request(url=api_url, headers={"Authorization": "Bearer " + kube_token}, is_json=True).json()

            # extract the hosts ip addresses
            ip_addresses = res.get("brothers")
            if api_res and isinstance(api_res, dict) and api_res.get("subsets") \
                    and isinstance(api_res.get("subsets"), list):
                subsets = api_res.get("subsets")
                for subset in subsets:
                    if subset.get("addresses") and isinstance(subset.get("addresses"), list):
                        addresses = subset.get("addresses")
                        for address in addresses:
                            host_ip = address.get("ip")
                            # if this host entry is not in the ignore_list, then add it
                            if host_ip and host_ip not in ignore_list:
                                ip_addresses.append(host_ip)
            else:
                raise Exception("Invalid Kubernetes API response")
        else:  # platform == "docker":
            # ".[].NetworkSettings.Networks.mike.IPAddress"
            #
            # curl --unix-socket /var/run/docker.sock -X GET \
            # http://v2/containers/json?filters=%7B%22label%22%3A%5B%22com.docker.swarm.service.name%3Dmy_service%22%5D%7D
            ip_addresses = res.get("brothers")
            client = docker.from_env()
            label_filter = "com.docker.swarm.service.name=" + self._configuration.get("platformOptions").get("service")
            network_name = self._configuration.get("platformOptions").get("network")
            containers = client.containers.list(filters={"label": [label_filter]})
            # loop over containers to get their IPs
            try:
                for cont in containers:
                    if cont.attrs["NetworkSettings"] and cont.attrs["NetworkSettings"]["Networks"] \
                            and cont.attrs["NetworkSettings"]["Networks"] \
                            and cont.attrs["NetworkSettings"]["Networks"][network_name]:
                        host_ip = cont.attrs["NetworkSettings"]["Networks"][network_name]["IPAddress"]
                        # if this host entry is not in the ignore_list, then add it
                        if host_ip and host_ip not in ignore_list:
                            ip_addresses.append(host_ip)
            except Exception:
                _logger.error(traceback.format_exc())

        return res

    def notify_brothers(self, endpoint, ignore_list):
        res = self.get_brothers(ignore_list)
        bros = res.get("brothers")
        upd_ignore_list = res.get("ignoreList")
        maintenance_port = self._configuration.get("contract").get("ports").get("maintenance")
        if bros and len(bros) > 0:
            for bro in bros:
                url = "https://" + bro + ":" + str(maintenance_port) + endpoint
                try:
                    utils.request(url=url, query={"ignoreList": upd_ignore_list}, is_json=True)
                except Exception:
                    _logger.error(traceback.format_exc())
                    _logger.error("Call failed to: " + url)



import pytest

from m360.awareness import instance
from m360.utils import from_json
from .helper import Request, random_int, random_string, random_numeric_string

awareness = instance()

def random_user():
    return {
        "id": random_int(1000, 9999),
        "firstname": random_string(15),
        "lastname": random_string(15, uppercase=True),
        "phone": random_numeric_string(8),
        "m360": {
            "m360_prop": "test"
        },
        "tenants": [{
            "id": random_int(1000, 9999),
            "code": "TENANT1"
        }, {
            "id": random_int(1000, 9999),
            "code": "TENANT2"
        }],
        "config": {
            "TENANT1": {"a": 1, "b": 2},
            "TENANT2": {"a": 3, "b": 4}
        }
    }

def random_user_with_config():
    usr = {
        "id": random_int(1000, 9999),
        "email": random_string(15) + "@" + random_string(8) + ".com",
        "firstname": random_string(15),
        "lastname": random_string(15, uppercase=True),
        "status": "active",
        "profile": {"gender": "male", "picture": ""},
        "ts": 1614693465902,
        "account": {
            "_id": "6041ed8e08acbf71381e0789",
            "name": "Corsair M360 - Henotic",
            "ts": 1614613159873,
        },
        "groups": ["admin"],
        "tenants": [{
            "id": "5f2abc6d3b62d4a05c929df9",
            "code": "M360"
        }],
        "config": {
            "common": {"foo": "bar"},
            "specific": {"test": {"foo": "bar"}}
        },
        "auto_registration_flow": "corsair_saas",
        "m360": {}
    }
    usr["m360"]["user"] = {
        "id": usr.get("id"),
        "username": usr.get("username"),
        "firstName": usr.get("firstName"),
        "lastName": usr.get("lastName"),
        "email": usr.get("email"),
        "status": usr.get("status"),
        "profile": usr.get("profile"),
        "ts": usr.get("ts"),
        "account": usr.get("account"),
        "groups": usr.get("groups"),
        "auto_registration_flow": usr.get("auto_registration_flow"),
        "m360": {}
    }
    usr["m360"]["tenants"] = usr.get("tenants"),
    return usr

@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before each test
    awareness._cache.clear()  # clear cache so that individual tests don't overlap
    # A test function will be run at this point
    yield
    # Code that will run after each test

def test_get_1(requests_mock):
    requests_mock.get("https://127.0.0.1:5000/awareness", json={"success": True})
    res = awareness.get()
    assert res.get("success")

def test_get_2(requests_mock):
    requests_mock.get("https://10.0.0.10:4000/awareness", json={"result": True, "data": {"success": True}})
    res = awareness.get("10.0.0.10", 4000)
    assert res.get("success")

def test_get_3(requests_mock):
    awareness.init(0.00001)  # 1 nanosecond i.e. no cache
    requests_mock.get("https://10.0.0.10:4000/awareness", json={"result": True, "data": {"test": 123}})
    res = awareness.get("10.0.0.10", 4000)
    assert res.get("test") == 123
    requests_mock.get("https://10.0.0.10:4000/awareness", json={"result": True, "data": {"test": 456}})
    res = awareness.get("10.0.0.10", 4000)
    assert res.get("test") == 456
    awareness.init()  # restore to default

def test_reload_1(requests_mock):
    url = "https://127.0.0.1:5000/awareness"
    requests_mock.get(url, json={"test": 123})
    res = awareness.get()
    assert res.get("test") == 123
    requests_mock.get(url, json={"test": 456})
    res = awareness.get()
    assert res.get("test") == 123  # cached value returned
    res = awareness.reload()
    assert res.get("test") == 456  # updated value returned
    res = awareness.get()
    assert res.get("test") == 456  # new cached value returned

def test_reload_2(requests_mock):
    url = "https://10.0.1.10:4000/awareness"
    requests_mock.get(url, json={"test": "ok"})
    res = awareness.get("10.0.1.10", 4000)
    assert res.get("test") == "ok"
    requests_mock.get(url, json={"test": "not ok"})
    res = awareness.get("10.0.1.10", 4000)
    assert res.get("test") == "ok"  # cached value returned
    res = awareness.reload("10.0.1.10", 4000)
    assert res.get("test") == "not ok"  # updated value returned
    res = awareness.get("10.0.1.10", 4000)
    assert res.get("test") == "not ok"  # new cached value returned

def test_get_next_host_1(requests_mock):
    requests_mock.get("https://127.0.0.1:5000/service/nextHost", json={"success": True})
    res = awareness.get_next_host("my_service", 2)
    assert res.get("success")

def test_get_next_host_2(requests_mock):
    requests_mock.get("https://10.0.0.10:4000/service/nextHost", json={"result": True, "data": {"success": True}})
    res = awareness.get_next_host("another_service", 1, "10.0.0.10", 4000)
    assert res.get("success")

def test_proxy_1(requests_mock):
    requests_mock.get("https://10.0.0.10:4000/service/nextHost",
                      json={"result": True, "data": {"ip": "10.11.12.13", "port": 1234}})
    requests_mock.get("https://10.11.12.13:1234/test/route", json={"my_data": {"ok": True}})
    res = awareness.proxy({"service": "another_service", "version": 1, "gatewayIp": "10.0.0.10", "gatewayPort": 4000,
                           "route": "/test/route", "qs": {"param1": 123, "param2": "yes"},
                           "headers": {"MY_HEADER": "some-data-goes-here"}})
    assert res.get("my_data").get("ok")

def test_proxy_2(requests_mock):
    requests_mock.get("https://10.0.0.10:4000/service/nextHost",
                      json={"result": True, "data": {"ip": "10.11.12.13", "port": 1234}})
    requests_mock.post("https://10.11.12.13:1234/test/post", json={"my_data": {"ok": True}})
    res = awareness.proxy({"service": "one_service", "version": 1, "gatewayIp": "10.0.0.10", "gatewayPort": 4000,
                           "method": "post", "route": "/test/post", "body": {"param1": 123, "param2": "yes"},
                           "headers": {"MY_HEADER": "some-data-goes-here"}})
    assert res.get("my_data").get("ok")

def test_proxy_3(requests_mock):
    requests_mock.get("https://10.0.0.10:4000/service/nextHost",
                      json={"result": True, "data": {}})
    requests_mock.delete("https://127.0.0.1:5000/test/delete", json={"my_data": {"ok": True}})
    res = awareness.proxy({"service": "one_service", "version": 1, "gatewayIp": "10.0.0.10", "gatewayPort": 4000,
                           "method": "delete", "route": "/test/delete"})
    assert res.get("my_data").get("ok")

def test_proxy_4(requests_mock):
    requests_mock.get("https://10.0.0.10:4000/service/nextHost",
                      json={"result": True, "data": {"ip": "10.11.12.13", "port": 1234}})
    requests_mock.get("https://10.11.12.13:1234/test/route", json={"my_data": {"ok": True}})
    usr = random_user()
    requests_mock.get("https://127.0.0.1:5000/users/" + str(usr.get("id")), json={"result": True, "data": usr})
    res = awareness.proxy({"service": "another_service", "version": 1, "gatewayIp": "10.0.0.10", "gatewayPort": 4000,
                           "route": "/test/route", "qs": {"param1": 123, "param2": "yes"}, "user": usr.get("id"),
                           "headers": {"MY_HEADER": "some-data-goes-here"}})
    assert res.get("my_data").get("ok")

def test_proxy_5(requests_mock):
    requests_mock.get("https://10.0.0.10:4000/service/nextHost",
                      json={"result": True, "data": {"ip": "10.11.12.13", "port": 1234}})
    requests_mock.get("https://10.11.12.13:1234/test/route", json={"my_data": {"ok": True}})
    usr = random_user_with_config()
    requests_mock.get("https://127.0.0.1:5000/users/" + str(usr.get("id")), json={"result": True, "data": usr})
    context = {"service": "test", "version": 1, "gatewayIp": "10.0.0.10", "gatewayPort": 4000,
               "route": "/test/route", "qs": {"param1": 123, "param2": "yes"}, "user": usr.get("id"),
               "headers": {"MY_HEADER": "some-data-goes-here"}}
    res = awareness.proxy(context)
    assert res.get("my_data").get("ok")
    assert context.get("headers")
    assert context.get("headers").get("m360")
    m360_headers = from_json(context.get("headers").get("m360"))
    assert m360_headers.get("user")
    assert m360_headers.get("tenants")

def test_proxy_error_1(requests_mock):
    requests_mock.get("https://10.0.0.10:4000/service/nextHost",
                      json={"result": True, "data": {"ip": "10.11.12.13", "port": 1234}})
    requests_mock.get("https://10.11.12.13:1234/test/route", json={"my_data": {"ok": True}})
    requests_mock.get("https://127.0.0.1:5000/users/invalid-id", json={"result": True, "data": None})
    with pytest.raises(Exception) as e:
        awareness.proxy({"service": "another_service", "version": 1, "gatewayIp": "10.0.0.10", "gatewayPort": 4000,
                         "route": "/test/route", "qs": {"param1": 123, "param2": "yes"}, "user": "invalid-id",
                         "headers": {"MY_HEADER": "some-data-goes-here"}})
    assert "User Not Found!" in str(e)



from m360.gateway.connector import instance

connector = instance()

def test_invoke_1(requests_mock):
    requests_mock.get("https://127.0.0.1:5000", json={"success": True})
    res = connector.invoke()
    assert res.get("success")

def test_invoke_2(requests_mock):
    requests_mock.get("https://10.0.0.1:4000", json={"result": True, "data": {"success": True}})
    res = connector.invoke({"ip": "10.0.0.1", "port": 4000})
    assert res.get("success")

def test_invoke_3(requests_mock):
    connector.init("my_service", "1.2.3")
    requests_mock.post("https://10.0.0.1:5000/test/route", json={"result": True, "data": {"test": {"success": True}}})
    res = connector.invoke({"ip": "http://10.0.0.1//", "route": "/test/route", "method": "post",
                            "headers": {"XXX": "123"}, "qs": {"q1": "val1", "q2": 2},
                            "body": {"obj": {"name": "hg", "active": True, "details": {"address": "BEY"}}}})
    assert res.get("test").get("success")



import pytest

from m360 import utils

def test_json_file_to_dict_fail_empty_path():
    with pytest.raises(Exception) as e:
        utils.json_file_to_dict("")
    assert e

def test_is_string_success():
    assert utils.is_string("")
    assert utils.is_string("   ")
    assert utils.is_string("yes")
    assert not utils.is_string(1)
    assert not utils.is_string(True)
    assert not utils.is_string(False)
    assert not utils.is_string({})
    assert not utils.is_string([])

def test_remove_object_from_path_success():
    data = {"config": {"M360": {"common": {"primary": True, "pet": "cersei", "foo": "bar"}, "specific": {}}}}
    utils.remove_object_from_path(data, "config.M360.invalid.primary")
    assert data == {"config": {"M360": {"common": {"primary": True, "pet": "cersei", "foo": "bar"}, "specific": {}}}}
    utils.remove_object_from_path(data, "config.M360.common.primary")
    assert data == {"config": {"M360": {"common": {"pet": "cersei", "foo": "bar"}, "specific": {}}}}
    utils.remove_object_from_path(data, "config.M360")
    assert data == {"config": {}}



"""
Copyright (C) 2022 Corsair M360, Inc - All Rights Reserved.
Unauthorized copying of this file, via any medium is strictly prohibited.
Proprietary and confidential.
"""

import logging
from django.http import JsonResponse

from ..helper import helper_instance

_logger = logging.getLogger(__name__)

class DjangoMiddleware:
    """
    Django M360 Middleware
    """

    def __init__(self, get_response):
        from django.conf import settings
        from .settings import CONFIG

        self.get_response = get_response
        self.helper = helper_instance("django", CONFIG, getattr(settings, "M360", None))

    def __call__(self, request):
        # handle reserved routes
        data = self.helper.handle_request(request)
        if data:
            if data.get("name"):
                return JsonResponse(data, status=200)
            else:
                return JsonResponse(data, status=500)

        # everything above this line happens before the app's request handler execution
        response = self.get_response(request)  # handler execution
        # everything below here happens after the app's request handler execution

        return response



"""
Copyright (C) 2022 Corsair M360, Inc - All Rights Reserved.
Unauthorized copying of this file, via any medium is strictly prohibited.
Proprietary and confidential.
"""

import os

GATEWAY_MAINTENANCE_IP = os.environ["GATEWAY_MAINTENANCE_IP"] if os.environ.get("GATEWAY_MAINTENANCE_IP") \
    else "127.0.0.1"
GATEWAY_MAINTENANCE_PORT = int(os.environ["GATEWAY_MAINTENANCE_PORT"]) if os.environ.get("GATEWAY_MAINTENANCE_PORT") \
    else 5000

SVC_ACCOUNT_TOKEN_PATH = "/var/run/secrets/kubernetes.io/serviceaccount/token"

RESERVED_ROUTES = ["/heartbeat", "/registry/reload", "/awareness/reload"]

APP_AUTOREGISTER = True if os.environ.get("APP_AUTOREGISTER") is None or \
                           os.environ.get("APP_AUTOREGISTER").lower() in ["true", "yes", "1"] else False

LOG_LEVEL = os.environ.get("LOG_LEVEL", "INFO").upper()
MW_TESTING = True if os.environ.get("MW_TESTING") else False

DEFAULT_AWARENESS_TTL_MS = 3600000  # 1 hour
DEFAULT_RETRY_DELAY_SEC = 2  # 2 seconds





"""
Copyright (C) 2022 Corsair M360, Inc - All Rights Reserved.
Unauthorized copying of this file, via any medium is strictly prohibited.
Proprietary and confidential.
"""

import os
import logging
import threading

from m360 import sdk, constants
from m360.maintenance import Maintenance

_logger = logging.getLogger(__name__)

class Helper:
    """
    M360 Middleware Helper
    """

    # singleton
    _instance = None

    def __init__(self, framework, default_config, config=None):
        _logger.info("M360 middleware config: " + str(config))
        self.framework = framework
        self.config = self.init_config(default_config, config)
        _logger.info("M360 middleware merged config: " + str(self.config))

        # M360 SDK initialization
        self.sdk = sdk.instance()
        self.sdk.init(self.config)
        self.sdk.augment(self.config)
        self.app_name = self.config.get("contract", {}).get("name", "noname")
        self.maintenance = Maintenance(self.config, self.app_name, self.sdk.registry, self.sdk.service,
                                       self.sdk.awareness)
        self.init_done = False
        self.lazy_init_ttl = 0.1
        self.lazy_init_failures = 0

        threading.Timer(self.lazy_init_ttl, self.lazy_init).start()  # schedule the lazy_init method in 100 milliseconds
        _logger.info("M360 middleware loaded")

    def handle_request(self, request):
        _logger.debug("M360 middleware intercepted request: [" + request.method + "] " + request.path)
        query = request.GET if self.framework == "django" else request.args

        try:
            # check if this route is one of the reserved ones, if so handle it
            return self.handle_special_routes(request.path, query)
        except Exception as e:
            return {"error": str(e)}

    @staticmethod
    def singleton(framework, default_config, config=None):
        """Access the singleton"""

        if Helper._instance:
            return Helper._instance

        if not Helper._instance:
            _logger.debug("Creating middleware helper instance")
            Helper._instance = Helper(framework, default_config, config)

        return Helper._instance

    @staticmethod
    def init_config(default_config, config=None):
        """
        Builds the final config object by using defaults for non provided keys
        """
        if not config:
            config = default_config
            return config
        for k, v in default_config.items():
            if k not in config.keys():
                config.setdefault(k, v)
        return config

    def lazy_init(self):
        try:
            if self.init_done:
                return
            self.sdk.auto_register(self.config)
            self.init_done = True
            _logger.info("M360 middleware registered successfully")
        except Exception as e:
            self.lazy_init_failures += 1
            if self.lazy_init_failures <= 5:
                _logger.debug("M360 middleware registration failed, retrying in a bit...")
                self.lazy_init_ttl = self.lazy_init_ttl * 3
                threading.Timer(self.lazy_init_ttl, self.lazy_init).start()  # retry in a bit
            else:
                _logger.error(str(e))
                raise Exception("M360 middleware failed to register with the gateway")

    def handle_special_routes(self, path, query):
        if path not in constants.RESERVED_ROUTES:
            return None
        if path == "/heartbeat":
            return {
                "name": self.app_name,
                "heartbeat": True
            }
        if path == "/registry/reload":
            ignore_list = query.getlist("ignoreList") if query and query.getlist("ignoreList") else []
            return self.maintenance.reload_registry(ignore_list)
        if path == "/awareness/reload":
            ignore_list = query.getlist("ignoreList") if query and query.getlist("ignoreList") else []
            return self.maintenance.reload_awareness(ignore_list)
        return None

def helper_instance(framework, default_config, config=None):
    """Get the singleton instance of the registry"""
    _logger.debug("calling Helper instance, os.pid=%d", os.getpid())
    return Helper.singleton(framework, default_config, config)



import pytest

from m360.utils import to_json
from m360.tenant import instance
from .helper import Request, random_int, random_string, random_numeric_string

tenant = instance()

def random_tenant():
    return {
        "id": random_int(1000, 9999),
        "code": random_string(8, uppercase=True),
        "name": random_string(15),
        "description": random_string(50),
        "phone": random_numeric_string(8)
    }

def random_tenants(count=random_int(3, 6)):
    arr = []
    for i in range(count):
        arr.append(random_tenant())
    return arr

def test_get():
    count = random_int(4, 7)
    ten = random_tenants(count)
    req = Request(headers={"m360": to_json({"tenants": ten}), "another-header": "--Some Value--"})

    for i in range(count):
        res = tenant.get(req, ten[i].get("code"))
        assert res == ten[i]

def test_list():
    ten = random_tenants()
    req = Request(headers={"m360": to_json({"tenants": ten}), "another-header": "--Some Value--"})

    res = tenant.list(req)
    assert res == ten

def test_find_one(requests_mock):
    ten = random_tenant()
    requests_mock.post("https://127.0.0.1:5000/tenants", json={"result": True, "data": ten})

    res = tenant.find([ten.get("code")])
    assert res == ten

def test_find_many(requests_mock):
    count = random_int(4, 7)
    ten = random_tenants(count)
    requests_mock.post("https://127.0.0.1:5000/tenants", json={"result": True, "data": ten})

    res = tenant.find()
    assert res == ten

def test_get_errors():
    count = random_int(4, 7)
    ten = random_tenants(count)
    req1 = Request(headers={"another-header": "--Some Value--"})
    req2 = Request(headers={"m360": '[]-{"invalid-json"}', "another-header": "--Some Value--"})
    req3 = Request(headers={"m360": to_json({"missing-tenants": ten}), "another-header": "--Some Value--"})
    req4 = Request(headers={"m360": to_json({"tenants": ten}), "another-header": "--Some Value--"})

    assert not tenant.get(req1, ten[0].get("code"))
    assert not tenant.get(req2, ten[0].get("code"))
    assert not tenant.get(req3, ten[0].get("code"))
    assert not tenant.get(req4, "inv-code")

def test_list_errors():
    req1 = Request(headers={"another-header": "--Some Value--"})
    req2 = Request(headers={"m360": '[]-{"invalid-json"}', "another-header": "--Some Value--"})
    req3 = Request(headers={"m360": to_json({"missing-tenants": random_tenants()}), "another-header": "--Some Value--"})
    req4 = Request(headers={"m360": to_json({"tenants": random_tenant()}), "another-header": "--Some Value--"})

    assert not tenant.list(req1)
    assert not tenant.list(req2)
    assert not tenant.list(req3)
    assert not tenant.list(req4)

def test_find_errors():
    with pytest.raises(Exception) as e:
        tenant.find()
    assert "Connection refused" in str(e)



"""
Copyright (C) 2022 Corsair M360, Inc - All Rights Reserved.
Unauthorized copying of this file, via any medium is strictly prohibited.
Proprietary and confidential.
"""

import os
import logging
import threading

from m360 import constants
from m360.gateway import connector
from m360.ttlcache import TTLCache

_logger = logging.getLogger(__name__)
_connector = connector.instance()
_cache = TTLCache(ttl=300)  # TTL defaults to 5 minutes

class Registry(object):
    """
    Registry class that fetches, caches and reloads the registry information of the environment by invoking the
    maintenance gateway API
    """

    # singleton
    _instance = None

    _registry_reload_timer = None

    def __init__(self):
        _logger.debug("Registry created")

    @staticmethod
    def registry_singleton():
        """Access the singleton receiver"""

        if Registry._instance:
            return Registry._instance

        if not Registry._instance:
            _logger.debug("Creating registry instance")
            Registry._instance = Registry()

        return Registry._instance

    @classmethod
    def __get_registry_section(cls, registry, section=None):
        """
        Returns registry[section] if it's there or registry if it isn't
        @param registry {dict}
        @param section  {str}
        @returns {dict} The specific registry section
        """
        if not registry:
            return None
        if section and type(section) is str and section.strip() and registry.get(section):
            return registry.get(section)
        return registry

    @classmethod
    def get(cls, section=None):
        """
        Returns the cached registry data
        If the cache is empty, it calls load, then it returns the result
        If the parameter is provided, it returns the specified portion of the registry
        @param section {str}
        @returns {dict} The registry record
        """
        registry = _cache.get("registry")
        if registry:
            return cls.__get_registry_section(registry, section)
        # load registry to continue
        cls.load()
        registry = _cache.get("registry")
        return cls.__get_registry_section(registry, section)

    @classmethod
    def clear(cls):
        """
        Clears the whole cache
        """
        _cache.clear()

    @staticmethod
    def load():
        """
        Calls the maintenance gateway API and caches the retrieved registry
        """
        delay = 0.001 if constants.MW_TESTING else constants.DEFAULT_RETRY_DELAY_SEC
        data = _connector.invoke({
            "method": "get",
            "route": "/registry/",
            "retryOnFailure": 3,
            "retryDelay": delay
        })
        # cache the retrieved info
        if data and isinstance(data, dict) and data.get("registry"):
            ttl = data.get("registry").get("_TTL") / 1000
            _cache.set("registry", data.get("registry"), ttl)
            _cache.set("custom", data.get("custom"), ttl)
            _cache.set("resources", data.get("resources"), ttl)
            _cache.set("databases", data.get("databases"), ttl)

        return True

    @classmethod
    def auto_reload(cls, ttl):
        """
        Starts a thread that handles reloading the registry periodically
        @param ttl {int}    Time to live in seconds, time between reloads
        """
        registry = _cache.get("registry")
        if registry and registry.get("_TTL"):
            ttl = registry.get("_TTL") / 1000
        _logger.info("Next registry reload in " + str(ttl) + " seconds...")

        def _reload():
            """Runs when the timer expires"""
            cls.load()
            cls.auto_reload(ttl)

        # cancel any running timers
        cls.stop_auto_reload()
        # schedule the _reload method in TTL seconds
        cls._registry_reload_timer = threading.Timer(ttl, _reload)
        cls._registry_reload_timer.start()

    @classmethod
    def reload(cls):
        """
        Calls load() and returns the latest registry
        """
        cls.load()
        return _cache.get("registry")

    @classmethod
    def stop_auto_reload(cls):
        """
        Stops any running timers
        """
        try:
            if cls._registry_reload_timer:
                cls._registry_reload_timer.cancel()
        except Exception:
            pass

class CustomRegistry(object):
    """Offers custom registry related helper methods"""

    # singleton
    _instance = None

    def __init__(self):
        _logger.debug("Custom registry created")

    @staticmethod
    def custom_registry_singleton():
        """Access the singleton receiver"""

        if CustomRegistry._instance:
            return CustomRegistry._instance

        if not CustomRegistry._instance:
            _logger.debug("Creating custom registry instance")
            CustomRegistry._instance = CustomRegistry()

        return CustomRegistry._instance

    @classmethod
    def get(cls, name=None):
        """
        Returns the requested custom registry based on its name
        @param name {str}
        @returns {dict} The custom_registry record
        """
        custom_registries = _cache.get("custom")
        if custom_registries and type(custom_registries) is dict and len(custom_registries) > 0:
            if name and name.strip():
                return custom_registries.get(name)
            else:
                return custom_registries
        return None

class Resource(object):
    """Offers resource related helper methods"""

    # singleton
    _instance = None

    def __init__(self):
        _logger.debug("Resource created")

    @staticmethod
    def resource_singleton():
        """Access the singleton receiver"""

        if Resource._instance:
            return Resource._instance

        if not Resource._instance:
            _logger.debug("Creating resource instance")
            Resource._instance = Resource()

        return Resource._instance

    @classmethod
    def get(cls, name=None):
        """
        Returns the requested resource based on its name
        @param name {str}
        @returns {dict} The resource record
        """
        resources = _cache.get("resources")
        if resources and type(resources) is dict and len(resources) > 0:
            if name and name.strip():
                return resources.get(name)
            else:
                return resources
        return None

class Database(object):
    """Offers database related helper methods"""

    # singleton
    _instance = None

    def __init__(self):
        _logger.debug("Database created")

    @staticmethod
    def database_singleton():
        """Access the singleton receiver"""

        if Database._instance:
            return Database._instance

        if not Database._instance:
            _logger.debug("Creating database instance")
            Database._instance = Database()

        return Database._instance

    @classmethod
    def __fix_db_name(cls, db_type, main, prefixed, tenant_code):
        """
        Format the DB name based on whether it's a single tenant or multi-tenant DB
        """
        if db_type == "multitenant":
            prefix = prefixed.replace(main, "") if main else prefixed
            if prefix == prefixed:
                if main:
                    return tenant_code + "_" + main
                else:
                    return tenant_code + "_" + prefix
            else:
                return prefix + tenant_code + "_" + main
        else:
            return prefixed

    @classmethod
    def __get_db(cls, db_type, name, tenant_code):
        """
        Load the specific database info from the cache by name
        """
        databases = _cache.get("databases")
        if databases and type(databases) is dict and databases.get(db_type):
            if name and name.strip() and databases.get(db_type).get(name):
                db = databases.get(db_type).get(name)
                cluster_name = db.get("cluster")
                db["cluster"] = resource_instance().get(cluster_name)
                if not db.get("cluster"):
                    db["cluster"] = {}
                db["cluster"]["name"] = cluster_name
                db["name"] = cls.__fix_db_name(db_type, name, db.get("name"), tenant_code)
                return db
            else:
                for db_name in databases.get(db_type):
                    cluster_name = databases.get(db_type).get(db_name).get("cluster")
                    databases[db_type][db_name]["cluster"] = resource_instance().get(cluster_name)
                    if not databases[db_type][db_name]["cluster"]:
                        databases[db_type][db_name]["cluster"] = {}
                    databases[db_type][db_name]["cluster"]["name"] = cluster_name
                    inner_db_name = databases.get(db_type).get(db_name).get("name")
                    databases[db_type][db_name]["name"] = cls.__fix_db_name(db_type, name, inner_db_name, tenant_code)
                return databases.get(db_type)
        return None

    @classmethod
    def get(cls, db_type, name=None, tenant_code=None):
        """
        Returns the requested database based on its name and db_type.
        If the database is multi-tenant, the tenant_code parameter is required to populate the DB name.
        @param db_type      {str}
        @param name         {str}
        @param tenant_code  {str}
        @returns {dict} The database record
        """
        if db_type == "single":
            return cls.__get_db("single", name, tenant_code)
        elif db_type == "multitenant":
            if not tenant_code or (type(tenant_code) is str and not tenant_code.strip()):
                return None
            else:
                return cls.__get_db("multitenant", name, tenant_code)
        else:
            return None

def registry_instance():
    """Get the singleton instance of the registry"""
    _logger.debug("calling registry_instance, os.pid=%d", os.getpid())
    return Registry.registry_singleton()

def custom_registry_instance():
    """Get the singleton instance of the custom_registry"""
    _logger.debug("calling custom_registry_instance, os.pid=%d", os.getpid())
    return CustomRegistry.custom_registry_singleton()

def resource_instance():
    """Get the singleton instance of the resource"""
    _logger.debug("calling resource_instance, os.pid=%d", os.getpid())
    return Resource.resource_singleton()

def database_instance():
    """Get the singleton instance of the database"""
    _logger.debug("calling database_instance, os.pid=%d", os.getpid())
    return Database.database_singleton()



"""
Copyright (C) 2022 Corsair M360, Inc - All Rights Reserved.
Unauthorized copying of this file, via any medium is strictly prohibited.
Proprietary and confidential.
"""

import logging
from werkzeug.wrappers import Request, Response

from ..helper import helper_instance
from .settings import CONFIG
from m360 import utils

_logger = logging.getLogger(__name__)

class FlaskMiddleware:
    """
    Flask M360 Middleware
    """

    def __init__(self, app, config):
        self.app = app
        self.helper = helper_instance("flask", CONFIG, config)

    def __call__(self, environ, start_response):
        request = Request(environ)
        # handle reserved routes
        data = self.helper.handle_request(request)
        if data:
            if data.get("name"):
                res = Response(utils.to_json(data), status=200, headers={"Content-Type": "application/json"})
                return res(environ, start_response)
            else:
                res = Response(utils.to_json(data), status=500, headers={"Content-Type": "application/json"})
                return res(environ, start_response)

        # continue regular handler execution
        return self.app(environ, start_response)



"""
Copyright (C) 2022 Corsair M360, Inc - All Rights Reserved.
Unauthorized copying of this file, via any medium is strictly prohibited.
Proprietary and confidential.
"""

import os
import logging
from jsonschema import RefResolver, Draft7Validator
from jsonschema.exceptions import ValidationError

from m360 import utils

DIR_NAME = os.path.dirname(__file__)
API_SCHEMA_PATH = "schemas/api.json"
CONTRACT_SCHEMA_PATH = "schemas/contract.json"
MIDDLEWARE_SCHEMA_PATH = "schemas/middleware.json"

_logger = logging.getLogger(__name__)

class Validator(object):
    """Offers validator related helper methods"""

    # singleton
    _instance = None

    def __init__(self):
        _logger.debug("Validator created")
        self.schemas = {
            "api": utils.json_file_to_dict(os.path.join(DIR_NAME, API_SCHEMA_PATH)),
            "contract": utils.json_file_to_dict(os.path.join(DIR_NAME, CONTRACT_SCHEMA_PATH)),
            "middleware": utils.json_file_to_dict(os.path.join(DIR_NAME, MIDDLEWARE_SCHEMA_PATH))
        }
        self.schema_store = {
            self.schemas["api"]["$id"]: self.schemas["api"],
            self.schemas["contract"]["$id"]: self.schemas["contract"],
            self.schemas["middleware"]["$id"]: self.schemas["middleware"]
        }
        self.ValidationError = ValidationError

    @staticmethod
    def validator_singleton():
        """Access the singleton receiver"""

        if Validator._instance:
            return Validator._instance

        if not Validator._instance:
            _logger.debug("Creating validator instance")
            Validator._instance = Validator()

        return Validator._instance

    def validate(self, data, schema_name):
        resolver = RefResolver.from_schema(self.schemas.get("api"), store=self.schema_store)
        validator = Draft7Validator(self.schemas.get(schema_name), resolver=resolver)
        return validator.validate(data)


def instance():
    """Get the singleton instance of the validator"""
    _logger.debug("calling instance, os.pid=%d", os.getpid())
    return Validator.validator_singleton()



"""
Copyright (C) 2022 Corsair M360, Inc - All Rights Reserved.
Unauthorized copying of this file, via any medium is strictly prohibited.
Proprietary and confidential.
"""

import os
import logging

from m360 import user, constants, utils
from m360.gateway import connector
from m360.ttlcache import TTLCache

_logger = logging.getLogger(__name__)
_connector = connector.instance()

def get_config(config, microservice):
    final_config = {}
    if config:
        # check and merge common config
        if config.get("common"):
            for key, val in config.get("common"):
                final_config[key] = val
        # check and merge specific microservice config
        if config.get("specific") and microservice and config.get("specific").get(microservice):
            temp_config = config.get("specific").get(microservice).get("config")
            if not temp_config:
                temp_config = config.get("specific").get(microservice)
            for key, val in temp_config:
                final_config[key] = val
            if config.get("specific").get(microservice).get("config") \
                    and config.get("specific").get(microservice).get("version"):
                final_config["__M360SV_" + microservice] = config.get("specific").get(microservice).get("version")
    # return merged config
    return final_config

class Awareness(object):
    """Offers awareness related helper methods"""

    # singleton
    _instance = None

    _ttl = 30
    _cache = TTLCache(ttl=_ttl)

    def __init__(self):
        _logger.debug("Awareness created")

    def init(self, ttl=None):
        if not ttl:
            ttl = constants.DEFAULT_AWARENESS_TTL_MS
        self._ttl = ttl / 1000
        _logger.debug("Awareness init complete")

    @staticmethod
    def awareness_singleton():
        """Access the singleton receiver"""

        if Awareness._instance:
            return Awareness._instance

        if not Awareness._instance:
            _logger.debug("Creating awareness instance")
            Awareness._instance = Awareness()

        return Awareness._instance

    def reload(self, ip=None, port=None, headers=None):
        self._cache.clear()
        return self.get(ip, port, headers)

    def get(self, ip=None, port=None, headers=None):
        no_cache = True if headers and headers.get("Cache-Control") == "no-cache" else False
        cache_key = str(ip) + "_" + str(port)
        cached = self._cache.get(cache_key)

        # return cached version if found
        if not no_cache and cached:
            return cached

        res = _connector.invoke({"ip": ip, "port": port, "headers": headers, "route": "/awareness"})
        # cache the value for next time
        self._cache.set(cache_key, res, self._ttl)
        return res

    @staticmethod
    def get_next_host(service, version, ip=None, port=None):
        qs = {"service": service, "version": version}
        return _connector.invoke({"ip": ip, "port": port, "route": "/service/nextHost", "qs": qs})

    @staticmethod
    def proxy(context):
        host = Awareness.get_next_host(context.get("service"), context.get("version"), context.get("gatewayIp"),
                                       context.get("gatewayPort"))

        if context.get("user"):
            user_doc = user.instance().find({"id": context.get("user")})

            if not user_doc:
                raise Exception("User Not Found!")

            if not context.get("headers"):
                context["headers"] = {}
            context["headers"]["m360"] = user_doc.get("m360")

            if user_doc.get("tenants"):
                context["headers"]["m360"]["tenants"] = []
                for t in user_doc.get("tenants"):
                    tenant = {
                        "id": t.get("id"),
                        "code": t.get("code"),
                        "config": get_config(user_doc.get("config").get(t.get("code")), context.get("service"))
                    }
                    context["headers"]["m360"]["tenants"].append(tenant)

            context["headers"]["m360"] = utils.to_json(context["headers"]["m360"])

        return _connector.invoke({"ip": host.get("ip"), "port": host.get("port"), "route": context.get("route"),
                                  "qs": context.get("qs"), "body": context.get("body"), "method": context.get("method"),
                                  "headers": context.get("headers")})

def instance():
    """Get the singleton instance of the awareness"""
    _logger.debug("calling instance, os.pid=%d", os.getpid())
    return Awareness.awareness_singleton()



"""
Copyright (C) 2022 Corsair M360, Inc - All Rights Reserved.
Unauthorized copying of this file, via any medium is strictly prohibited.
Proprietary and confidential.
"""

CONFIG = {
    # default config goes here
}



"""
Copyright (C) 2022 Corsair M360, Inc - All Rights Reserved.
Unauthorized copying of this file, via any medium is strictly prohibited.
Proprietary and confidential.
"""

import logging
import time
from copy import deepcopy

_logger = logging.getLogger(__name__)

class TTLCache(object):
    """In memory cache implementation with TTL support.
    Items are expired and popped during mutating operations instead of using periodic checks."""

    _cache = {}
    _ttl = 300  # in seconds, defaults to 5 minutes

    def __init__(self, ttl):
        self._ttl = ttl
        _logger.debug("TTLCache init complete")

    def _check_ttls(self):
        """
        Cycles through all cached items and pops the ones that are expired
        """
        now = time.time()
        to_pop = []
        # build a list of expired keys
        for key in self._cache:
            item = self._cache.get(key)
            if item.get("exp") < now:
                to_pop.append(key)
        # remove the expired keys from the cache
        for key in to_pop:
            self._cache.pop(key, None)

    def set(self, key, value, ttl=_ttl):
        """
        Set a key/value pair in the cache with a specific ttl in seconds (or the default one)
        @param key      {str}
        @param value    {any}
        @param ttl      {int}   Time to live in seconds
        """
        self._check_ttls()
        self._cache[key] = {"exp": time.time() + ttl, "value": value}

    def get(self, key):
        """
        Get a cached value using its key
        @param key      {str}
        @returns {any}
        """
        self._check_ttls()
        return deepcopy(self._cache.get(key).get("value")) if self._cache.get(key) else None

    def pop(self, key):
        """
        Remove an item from the cache using its key
        @param key      {str}
        """
        self._check_ttls()
        self._cache.pop(key, None)

    def clear(self):
        """
        Clears the cache i.e. removes all saved items
        """
        self._cache = {}



"""
Copyright (C) 2022 Corsair M360, Inc - All Rights Reserved.
Unauthorized copying of this file, via any medium is strictly prohibited.
Proprietary and confidential.
"""

import os
import json
import logging

from m360 import utils
from m360.gateway import connector

_logger = logging.getLogger(__name__)
_connector = connector.instance()

class Tenant(object):
    """Offers tenant related helper methods"""

    # singleton
    _instance = None

    def __init__(self):
        _logger.debug("Tenant created")

    @staticmethod
    def tenant_singleton():
        """Access the singleton receiver"""

        if Tenant._instance:
            return Tenant._instance

        if not Tenant._instance:
            _logger.debug("Creating tenant instance")
            Tenant._instance = Tenant()

        return Tenant._instance

    @staticmethod
    def get(request, code):
        if not code or not code.strip() or not request:
            return None
        headers = utils.get_req_headers(request)
        if not headers or "m360" not in headers:
            return None

        try:
            m360 = headers.get("m360")
            if m360:
                m360 = json.loads(m360)
        except Exception as e:
            _logger.error(e)
            m360 = None

        if not m360 or "tenants" not in m360 or type(m360.get("tenants")) is not list:
            _logger.error("Invalid schema for m360=" + str(m360))
            return None

        tenant = None
        for t in m360.get("tenants"):
            if t and t.get("code") == code:
                tenant = t
                break
        return tenant

    @staticmethod
    def list(request):
        if not request:
            return None
        headers = utils.get_req_headers(request)
        if not headers or not headers.get("m360"):
            return None

        try:
            m360 = headers.get("m360")
            if m360:
                m360 = json.loads(m360)
        except Exception as e:
            _logger.error(e)
            m360 = None

        if not m360 or not m360.get("tenants") or type(m360.get("tenants")) is not list:
            _logger.error("Invalid schema for m360=" + str(m360))
            return None

        return m360.get("tenants")

    @staticmethod
    def find(codes=None):
        code = codes if codes else []
        url = "/tenants"
        body = {"code": code}

        return _connector.invoke({"route": url, "method": "post", "body": body})

def instance():
    """Get the singleton instance of the tenant"""
    _logger.debug("calling tenant_instance, os.pid=%d", os.getpid())
    return Tenant.tenant_singleton()



import pytest

from m360.utils import to_json
from m360.user import instance
from .helper import Request, random_int, random_string, random_numeric_string

user = instance()

def random_user():
    return {
        "id": random_int(1000, 9999),
        "firstname": random_string(15),
        "lastname": random_string(15, uppercase=True),
        "phone": random_numeric_string(8)
    }

def random_users(count=random_int(3, 6)):
    arr = []
    for i in range(count):
        arr.append(random_user())
    return arr

def test_get():
    usr = random_user()
    req = Request(headers={"m360": to_json({"user": usr}), "another-header": "--Some Value--"})

    res = user.get(req)
    assert res == usr

def test_find_one(requests_mock):
    usr = random_user()
    requests_mock.get("https://127.0.0.1:5000/users/" + str(usr.get("id")), json={"result": True, "data": usr})

    res = user.find({"id": usr.get("id"), "tenant": "XYZ", "group": "GRP1", "service": "aaa", "version": 1})
    assert res == usr

def test_find_many(requests_mock):
    count = random_int(4, 7)
    usr = random_users(count)
    requests_mock.get("https://127.0.0.1:5000/users?start=2&limit=5", json={"result": True, "data": usr})

    res = user.find({"start": 2, "limit": 5})
    assert res == usr

def test_get_errors():
    count = random_int(4, 7)
    usr = random_users(count)
    req1 = Request(headers={"another-header": "--Some Value--"})
    req2 = Request(headers={"m360": '[]-{"invalid-json"}', "another-header": "--Some Value--"})
    req3 = Request(headers={"m360": to_json({"missing-users": usr}), "another-header": "--Some Value--"})
    req4 = Request(headers={"m360": to_json({"user": None}), "another-header": "--Some Value--"})

    assert not user.get(req1)
    assert not user.get(req2)
    assert not user.get(req3)
    assert not user.get(req4)

def test_find_errors():
    with pytest.raises(Exception) as e:
        user.find({})
    assert "Connection refused" in str(e)




"""
Copyright (C) 2022 Corsair M360, Inc - All Rights Reserved.
Unauthorized copying of this file, via any medium is strictly prohibited.
Proprietary and confidential.
"""

import os
import logging

from m360 import constants, user, rbac, tenant, service, registry, awareness, validator, utils
from m360.gateway import connector

logging.basicConfig(level=constants.LOG_LEVEL)
_logger = logging.getLogger(__name__)
_connector = connector.instance()

class SDK(object):
    """Offers sdk related helper methods"""

    # singleton
    _instance = None

    def __init__(self):
        # set the logger
        self.logger = logging.getLogger("M360")
        # expose SDK libraries
        self.validator = validator.instance()
        self.awareness = awareness.instance()
        self.user = user.instance()
        self.rbac = rbac.instance()
        self.tenant = tenant.instance()
        self.service = service.instance()
        self.registry = registry.registry_instance()
        self.database = registry.database_instance()
        self.resource = registry.resource_instance()
        self.custom = registry.custom_registry_instance()
        self.app_name = None
        self.app_type = None
        self.config = None
        _logger.debug("SDK created")

    @staticmethod
    def sdk_singleton():
        """Access the singleton receiver"""

        if SDK._instance:
            return SDK._instance

        if not SDK._instance:
            _logger.debug("Creating sdk instance")
            SDK._instance = SDK()

        return SDK._instance

    def init(self, config):
        """
        Initialize the middleware and validate the provided configuration
        @param config {dict}
        """
        _logger.info("Initializing M360 SDK with the following config:")
        _logger.info(str(config))
        # SSL checks
        if not os.environ.get("APP_SSL_KEY") or not os.environ.get("APP_SSL_CERT"):
            _logger.error("""
\nError: Missing Application Private SSL certificate/key pair.
Please provide the paths to both the SSL certificate and its key either using environment variables ('APP_SSL_CERT' and 'APP_SSL_KEY')
\n#########################################################################################
\nIf you don't have a certificate/key pair, you can use openssl to generate one.
Example:\n> openssl req -new -x509 -sha256 -nodes -days 365 -newkey rsa:2048 -keyout /user/privatekey.key -out /user/certificate.crt
\nReference: https://www.baeldung.com/openssl-self-signed-cert
\n#########################################################################################\n
            """)
            self.exit_app()

        try:
            # load contract if a path was specified
            if config.get("contract") and type(config.get("contract")) is str:
                config["contract"] = utils.json_file_to_dict(config.get("contract"))

            # validate config schema
            _logger.debug("Validating configuration schema")
            self.validator.validate(config, "middleware")
            _logger.debug("Configuration schema is valid")

            if config.get("platform") in ["docker", "kubernetes"]:
                if not config.get("platformOptions") or type(config.get("platformOptions")) is not dict:
                    raise Exception("Missing 'platformOptions' for platform " + config.get("platform") +
                                    ". Please refer to documentation for additional help")
                errors = []
                if config.get("platform") == "docker":
                    # 'network': 'mike',
                    # 'service': 'service-express',
                    if not config.get("platformOptions").get("network"):
                        errors.append("Missing docker 'network' name in platform options")
                    if not config.get("platformOptions").get("service"):
                        errors.append("Missing docker 'service' name in platform options")
                else:  # == kubernetes
                    # 'namespace': 'mike',
                    # 'service': 'service-express',
                    # 'exposedPort': os.environ["APP_PORT"] or 4002,
                    if not config.get("platformOptions").get("namespace"):
                        errors.append("Missing docker 'namespace' value in platform options")
                    if not config.get("platformOptions").get("service"):
                        errors.append("Missing docker 'service' name in platform options")
                    if not config.get("platformOptions").get("exposedPort"):
                        errors.append("Missing docker 'exposedPort' value for container in platform options")

                if len(errors) > 0:
                    raise Exception("\n".join(errors))

            # check if preserved routes are overridden
            for api_object in config.get("contract").get("apis").values():
                for method, api_route in api_object.items():
                    for route_path in api_route.keys():
                        if method.lower() == "get" and route_path in constants.RESERVED_ROUTES:
                            raise Exception("Route [GET] " + route_path + " is reserved by the middleware, " +
                                            "please use a different route or method for your endpoint")

            # save local reference to config
            self.config = config
        except Exception as e:
            _logger.error(e)
            self.exit_app()

    def init_job(self, config):
        """
        Initialize the different libraries
        @param config {dict}
        """
        config["type"] = "job"
        config["server"] = {}

        self.init(config)

        _logger.info("Retrieving Service Contract from M360 Gateway")
        self.service.get(config.get("contract"))  # we just make sure there are no exceptions here
        _logger.info("Service Found in M360 Gateway")

        # initialize the middleware sdk gateway connector driver
        _connector.init(config.get("contract").get("name"), config.get("contract").get("version"))

        # get the registry from the gateway and cache it
        registry_data = self.registry.get()

        # if the registry was loaded...
        if registry_data:
            # initialize the awareness sdk module
            self.awareness.init(registry_data.get("_TTL"))
            _logger.info("Connection to M360 Gateway established")
            _logger.info("Environment Registry loaded from M360 Gateway and cached")

        # auto-wire the registry reload procedure with a default TTL of 30 seconds
        self.registry.auto_reload(30)

    def augment(self, config):
        """
        Augment the default configuration of the service contract and add the M360 maintenance routes to it.
        Augment the default microservice apis and add the M360 cloud awareness routes to the server instance.
        @param config {dict}
        """
        self.app_name = config.get("contract").get("name")
        self.app_type = config.get("type")

        # update the service contract with preserved maintenance entries
        config["contract"]["heartbeat"] = {
            "method": "get",
            "route": "/heartbeat",
            "access": False,
            "label": "Heartbeat"
        }
        config["contract"]["registry"] = {
            "method": "get",
            "route": "/registry/reload",
            "access": False,
            "label": "Reload Registry"
        }

        if not config.get("contract").get("apis").get("maintenance"):
            config["contract"]["apis"]["maintenance"] = {}
        if not config.get("contract").get("apis").get("maintenance").get("get"):
            config["contract"]["apis"]["maintenance"]["get"] = {}

        config["contract"]["apis"]["maintenance"]["get"]["/awareness/reload"] = {
            "access": False,
            "label": "Reload Cached Awareness"
        }

        self.rbac.set_service_config(config)

    def auto_register(self, config):
        """
        Method that triggers the auto registration procedure and registers/updates the microservice in the gateway.
        Once registered, the gateway will be able to proxy requests to it.
        @param config {dict}
        """
        # auto register the service at the gateway unless there's an env var set to stop it
        if constants.APP_AUTOREGISTER:
            self.service.ensure_ip_address(config)

            _logger.info("Registering/Updating Service Contract in M360 Gateway")
            response = self.service.register(config.get("contract"), config.get("ip"), config)
            _logger.info("Service Registered in M360 Gateway")

            # initialize the middleware sdk gateway connector driver
            _connector.init(config.get("contract").get("name"), config.get("contract").get("version"))

            # get the registry from the gateway and cache it
            registry_data = self.registry.get()

            if registry_data:  # if the registry was loaded..
                # initialize the awareness sdk module
                self.awareness.init(registry_data.get("_TTL"))

                # display all registration messages to the terminal
                if response and response.get("notifications") and len(response.get("notifications")) > 0:
                    _logger.info("\n*********************")
                    _logger.info("Registration Details:")
                    _logger.info("*********************")
                    _logger.info(str(response.get("notifications")))

                _logger.info("Connection to M360 Gateway established")
                _logger.info("Environment Registry loaded from M360 Gateway and cached")

            # auto-wire the registry reload procedure with a default TTL of 30 seconds
            self.registry.auto_reload(30)

    def stop_auto_reload(self):
        """
        Stops automatically reloading the registry (mainly useful for testing)
        """
        self.registry.stop_auto_reload()


    @staticmethod
    def exit_app():
        _logger.fatal("Exiting due to a fatal error")
        quit()

def instance():
    """Get the singleton instance of the sdk"""
    _logger.debug("calling instance, os.pid=%d", os.getpid())
    return SDK.sdk_singleton()



import pytest

from m360.utils import to_json
from m360.user import instance as user_instance
from m360.rbac import instance as rbac_instance
from .helper import Request

user = user_instance()
rbac = rbac_instance()

def build_request(m360_header=None):
    if m360_header:
        return Request(headers={"m360": m360_header})

    return Request(headers={
        "m360": to_json({
            "API": {
                "method": "get",
                "endpoint": "/"
            },
            "user": {
                "id": "12345",
                "username": "owner",
                "email": "mike@corsairm360.com",
                "account": {
                    "_id": "6041ed8e08acbf71381e0789",
                    "name": "mikeorg",
                }
            },
            "rbac": {
                "fields": {
                    "operator": "allow",
                    "list": {
                        "first": "firstName",
                        "username": "username",
                        "account": "account",
                        "config": "config.M360.common.pet"
                    }
                },
                "resources": {
                    "mode": "own",
                    "field": "account",
                    "value": "account._id"
                },
                "conditions": {
                    "operator": "$and",
                    "criteria": [
                        {
                            "function": "EQ",
                            "arguments": {
                                "field": "username",
                                "value": "username",
                                "custom": ""
                            }
                        }, {
                            "function": "EQ",
                            "arguments": {
                                "field": "account",
                                "value": "account._id",
                                "custom": ""
                            }
                        }, {
                            "function": "EQ",
                            "arguments": {
                                "field": "username",
                                "value": "custom",
                                "custom": "owner"
                            }
                        }, {
                            "function": "EQ",
                            "arguments": {
                                "field": "config.M360.common.primary",
                                "value": "custom",
                                "custom": "true"
                            }
                        }
                    ]
                }
            }
        })
    })

def test_set_service_config_1():
    rbac.set_service_config({
        "contract": {
            "name": "express",
            "group": "Testing",
            "version": 1,
            "ports": {
                "data": 4002,
                "maintenance": 4002
            },
            "apis": {
                "main": {
                    "get": {
                        "/": {
                            "label": "get api",
                            "access": True,
                            "rbac": {
                                "fields": ["config"]
                            }
                        }
                    },
                    "post": {
                        "/": {
                            "access": False,
                            "label": "post api",
                            "rbac": {
                                "fields": ["config"]
                            }
                        }
                    }
                }
            }
        }
    })

def test_can_access_fail_invalid():
    with pytest.raises(Exception) as e:
        rbac.can_access(build_request(to_json({
            "API": {
                "method": "get",
                "endpoint": "/"
            },
            "rbac": {
                "fields": {}
            }
        })), "invalid", {})
    assert e

def test_get_fail_no_m360():
    rb = rbac.get(Request(headers={}))
    assert not rb

def test_get_fail_m360_not_json():
    rb = rbac.get(build_request("invalid"))
    assert not rb

def test_get_fail_m360_missing_rbac():
    rb = rbac.get(build_request("{}"))
    assert not rb

def test_get_success():
    request = build_request()
    rb = rbac.get(request)
    assert rb
    assert rb.get("fields")
    assert rb.get("resources")
    assert rb.get("conditions")

    rbac_options_1 = ["fields", "field"]
    rbac_options_2 = ["resources", "resource"]
    rbac_options_3 = ["conditions", "condition"]

    for option in rbac_options_1:
        fields = rbac.get(request, option)
        assert fields.get("operator")
        assert fields.get("list")
    for option in rbac_options_2:
        resources = rbac.get(request, option)
        assert resources.get("mode")
        assert resources.get("field")
        assert resources.get("value")
    for option in rbac_options_3:
        conditions = rbac.get(request, option)
        assert conditions.get("operator")
        assert conditions.get("criteria")

    with pytest.raises(Exception) as e:
        rbac.get(request, "invalid")
    assert e

def test_can_access_fields_fail_invalid_data():
    with pytest.raises(Exception) as e:
        rbac.can_access(build_request(to_json({
            "API": {
                "method": "get",
                "endpoint": "/"
            },
            "rbac": {
                "fields": {}
            }
        })), "fields", "string")
    assert e

def test_can_access_fields_success_no_rbac_configured():
    status = rbac.can_access(build_request(to_json({
        "API": {
            "method": "get",
            "endpoint": "/"
        },
        "rbac": {
            "fields": {
                "operator": None
            }
        }
    })), "fields", {"foo": "bar"})
    assert status

def test_can_access_fields_success():
    data = {
        "user": "john",
        "isAdmin": True,
        "age": 32,
        "patt": "^[a-z]+$",
        "c": {
            "c": {
                "a": "b"
            }
        },
        "d": [
            ["a", "b"],
            ["c", "d"]
        ],
        "e": [
            {"a": "b"}
        ],
        "f": {
            "abc": {
                "access": ["admin", "vip"]
            }
        },
        "g": {
            "array": [
                {
                    "a": "b"
                },
                {
                    "g": "e"
                }
            ]
        }
    }
    tests1 = [
        {"entry1": "user"},
        {"entry2": "c.c.a"},
        {"entry3": "d.[1].[0]"},
        {"entry3": "d.[0].[0]"},
        {"entry4": "e.[0].a"},
        {"entry5": "f.abc.access.[0]"},
        {"entry6": "g.array.[0].a"},
        {"entry6": "g.array.[1].g"}
    ]
    tests2 = [
        {"entry6": "g.array.[1].f"}
    ]
    tests3 = [
        {"entry6": "g.array.[1].g"}
    ]

    def go(operator, tests, status_check, option):
        options = {"operator": operator, "list": {}}
        for test in tests:
            options["list"] = test
            status = rbac.can_access(build_request(to_json({
                "API": {
                    "method": "get",
                    "endpoint": "/"
                },
                "rbac": {
                    "fields": options
                }
            })), option, data)
            assert status == status_check

    go("allow", tests1, True, "fields")
    go("deny", tests2, True, "field")
    go("deny", tests3, True, "field")

def test_can_access_resources_fail_invalid_data():
    with pytest.raises(Exception) as e:
        rbac.can_access(build_request(to_json({
            "API": {
                "method": "get",
                "endpoint": "/"
            },
            "rbac": {
                "resources": {}
            }
        })), "resources", "string")
    assert e

def test_can_access_resources_success_no_rbac_configured():
    status = rbac.can_access(build_request(to_json({
        "API": {
            "method": "get",
            "endpoint": "/"
        },
        "rbac": {
            "resources": {
                "mode": None
            }
        }
    })), "resources", {"foo": "bar"})
    assert status

def test_can_access_resources_fail_no_user():
    status = rbac.can_access(build_request(to_json({
        "API": {
            "method": "get",
            "endpoint": "/"
        },
        "rbac": {
            "resources": {
                "mode": "any"
            }
        }
    })), "resources", {"foo": "bar"})
    assert not status

def test_can_access_resources_success_any():
    status = rbac.can_access(build_request(to_json({
        "user": {
            "id": "12345",
            "username": "owner",
            "email": "mike@corsairm360.com",
            "account": {
                "_id": "6041ed8e08acbf71381e0789",
                "name": "mikeorg",
            }
        },
        "API": {
            "method": "get",
            "endpoint": "/"
        },
        "rbac": {
            "resources": {
                "mode": "any"
            }
        }
    })), "resources", {"foo": "bar"})
    assert status

def test_can_access_resources_success_own():
    status = rbac.can_access(build_request(to_json({
        "user": {
            "id": "12345",
            "username": "owner",
            "email": "mike@corsairm360.com",
            "account": {
                "_id": "6041ed8e08acbf71381e0789",
                "name": "mikeorg",
            }
        },
        "API": {
            "method": "get",
            "endpoint": "/"
        },
        "rbac": {
            "resources": {
                "mode": "own",
                "field": "account",
                "value": "account._id"
            }
        }
    })), "resource", {
        "id": "12345",
        "username": "owner",
        "email": "mike@corsairm360.com",
        "account": "6041ed8e08acbf71381e0789"
    })
    assert status

def test_can_access_conditions_fail_invalid_data():
    with pytest.raises(Exception) as e:
        rbac.can_access(build_request(to_json({
            "API": {
                "method": "get",
                "endpoint": "/"
            },
            "rbac": {
                "conditions": {}
            }
        })), "conditions", "string")
    assert e

def test_can_access_conditions_success_no_rbac_configured():
    status = rbac.can_access(build_request(to_json({
        "API": {
            "method": "get",
            "endpoint": "/"
        },
        "rbac": {
            "conditions": {
                "operator": "$and"
            }
        }
    })), "conditions", {"foo": "bar"})
    assert status

def test_can_access_conditions_fail_no_user():
    status = rbac.can_access(build_request(to_json({
        "API": {
            "method": "get",
            "endpoint": "/"
        },
        "rbac": {
            "conditions": {
                "operator": "$and",
                "criteria": [
                    {
                        "function": "EQ",
                        "arguments": {
                            "field": "username",
                            "value": "username",
                            "custom": ""
                        }
                    },
                    {
                        "function": "EQ",
                        "arguments": {
                            "field": "account",
                            "value": "account._id",
                            "custom": ""
                        }
                    },
                    {
                        "function": "EQ",
                        "arguments": {
                            "field": "username",
                            "value": "custom",
                            "custom": "owner"
                        }
                    },
                    {
                        "function": "EQ",
                        "arguments": {
                            "field": "config.M360.common.primary",
                            "value": "custom",
                            "custom": "true"
                        }
                    }
                ]
            }
        }
    })), "conditions", {"foo": "bar"})
    assert not status

def test_can_access_conditions_success_and():
    status = rbac.can_access(build_request(to_json({
        "user": {
            "id": "12345",
            "username": "owner",
            "email": "mike@corsairm360.com",
            "account": {
                "_id": "6041ed8e08acbf71381e0789",
                "name": "mikeorg",
            }
        },
        "API": {
            "method": "get",
            "endpoint": "/"
        },
        "rbac": {
            "conditions": {
                "operator": "$and",
                "criteria": [
                    {
                        "function": "EMPTY",
                        "arguments": {
                            "field": "password",
                            "value": "",
                            "custom": ""
                        }
                    },
                    {
                        "function": "NOT_EMPTY",
                        "arguments": {
                            "field": "username",
                            "value": "",
                            "custom": ""
                        }
                    },
                    {
                        "function": "EQ",
                        "arguments": {
                            "field": "account",
                            "value": "account._id",
                            "custom": ""
                        }
                    },
                    {
                        "function": "EQ",
                        "arguments": {
                            "field": "primary",
                            "value": "custom",
                            "custom": "true"
                        }
                    },
                    {
                        "function": "EQ",
                        "arguments": {
                            "field": "username",
                            "value": "custom",
                            "custom": "owner"
                        }
                    },
                    {
                        "function": "NOT_EQ",
                        "arguments": {
                            "field": "username",
                            "value": "custom",
                            "custom": "mike"
                        }
                    },
                    {
                        "function": "START",
                        "arguments": {
                            "field": "config.test",
                            "value": "custom",
                            "custom": "mi"
                        }
                    },
                    {
                        "function": "NOT_START",
                        "arguments": {
                            "field": "config.test",
                            "value": "custom",
                            "custom": "jj"
                        }
                    },
                    {
                        "function": "END",
                        "arguments": {
                            "field": "config.test",
                            "value": "custom",
                            "custom": "ke"
                        }
                    },
                    {
                        "function": "NOT_END",
                        "arguments": {
                            "field": "config.test",
                            "value": "custom",
                            "custom": "jj"
                        }
                    },
                    {
                        "function": "CONTAIN",
                        "arguments": {
                            "field": "config.test",
                            "value": "custom",
                            "custom": "ik"
                        }
                    },
                    {
                        "function": "NOT_CONTAIN",
                        "arguments": {
                            "field": "config.test",
                            "value": "custom",
                            "custom": "jj"
                        }
                    },
                    {
                        "function": "IN",
                        "arguments": {
                            "field": "config.test",
                            "value": "custom",
                            "custom": "ik"
                        }
                    },
                    {
                        "function": "NOT_IN",
                        "arguments": {
                            "field": "config.test",
                            "value": "custom",
                            "custom": "jj"
                        }
                    }
                ]
            }
        }
    })), "conditions", {
        "id": "12345",
        "username": "owner",
        "email": "mike@corsairm360.com",
        "account": "6041ed8e08acbf71381e0789",
        "primary": True,
        "config": {
            "test": "mike"
        }
    })
    assert status

def test_can_access_conditions_success_or():
    status = rbac.can_access(build_request(to_json({
        "user": {
            "id": "12345",
            "username": "owner",
            "email": "mike@corsairm360.com",
            "account": {
                "_id": "6041ed8e08acbf71381e0789",
                "name": "mikeorg",
            }
        },
        "API": {
            "method": "get",
            "endpoint": "/"
        },
        "rbac": {
            "conditions": {
                "operator": "$or",
                "criteria": [
                    {
                        "function": "EMPTY",
                        "arguments": {
                            "field": "password",
                            "value": "",
                            "custom": ""
                        }
                    },
                    {
                        "function": "EMPTY",
                        "arguments": {
                            "field": "username",
                            "value": "",
                            "custom": ""
                        }
                    }
                ]
            }
        }
    })), "conditions", {
        "id": "12345",
        "username": "owner",
        "email": "mike@corsairm360.com",
        "account": "6041ed8e08acbf71381e0789",
        "primary": True,
        "config": {
            "test": "mike"
        }
    })
    assert status

def test_can_access_conditions_fail_and():
    status = rbac.can_access(build_request(to_json({
        "user": {
            "id": "12345",
            "username": "owner",
            "email": "mike@corsairm360.com",
            "account": {
                "_id": "6041ed8e08acbf71381e0789",
                "name": "mikeorg",
            }
        },
        "API": {
            "method": "get",
            "endpoint": "/"
        },
        "rbac": {
            "conditions": {
                "operator": "$and",
                "criteria": [
                    {
                        "function": "EMPTY",
                        "arguments": {
                            "field": "username",
                            "value": "",
                            "custom": ""
                        }
                    },
                    {
                        "function": "NOT_EMPTY",
                        "arguments": {
                            "field": "password",
                            "value": "",
                            "custom": ""
                        }
                    },
                    {
                        "function": "EQ",
                        "arguments": {
                            "field": "account",
                            "value": "account.name",
                            "custom": ""
                        }
                    },
                    {
                        "function": "EQ",
                        "arguments": {
                            "field": "username",
                            "value": "custom",
                            "custom": "true"
                        }
                    },
                    {
                        "function": "EQ",
                        "arguments": {
                            "field": "email",
                            "value": "custom",
                            "custom": "owner"
                        }
                    },
                    {
                        "function": "NOT_EQ",
                        "arguments": {
                            "field": "username",
                            "value": "custom",
                            "custom": "owner"
                        }
                    },
                    {
                        "function": "START",
                        "arguments": {
                            "field": "config.test",
                            "value": "custom",
                            "custom": "jj"
                        }
                    },
                    {
                        "function": "NOT_START",
                        "arguments": {
                            "field": "config.test",
                            "value": "custom",
                            "custom": "mi"
                        }
                    },
                    {
                        "function": "END",
                        "arguments": {
                            "field": "config.test",
                            "value": "custom",
                            "custom": "jj"
                        }
                    },
                    {
                        "function": "NOT_END",
                        "arguments": {
                            "field": "config.test",
                            "value": "custom",
                            "custom": "ke"
                        }
                    },
                    {
                        "function": "CONTAIN",
                        "arguments": {
                            "field": "config.test",
                            "value": "custom",
                            "custom": "jj"
                        }
                    },
                    {
                        "function": "NOT_CONTAIN",
                        "arguments": {
                            "field": "config.test",
                            "value": "custom",
                            "custom": "ik"
                        }
                    },
                    {
                        "function": "IN",
                        "arguments": {
                            "field": "config.test",
                            "value": "custom",
                            "custom": "jj"
                        }
                    },
                    {
                        "function": "NOT_IN",
                        "arguments": {
                            "field": "config.test",
                            "value": "custom",
                            "custom": "ik"
                        }
                    }
                ]
            }
        }
    })), "conditions", {
        "id": "12345",
        "username": "owner",
        "email": "mike@corsairm360.com",
        "account": "6041ed8e08acbf71381e0789",
        "primary": True,
        "config": {
            "test": "mike"
        }
    })
    assert not status

def test_can_access_conditions_fail_invalid_condition():
    status = rbac.can_access(build_request(to_json({
        "user": {
            "id": "12345",
            "username": "owner",
            "email": "mike@corsairm360.com",
            "account": {
                "_id": "6041ed8e08acbf71381e0789",
                "name": "mikeorg",
            }
        },
        "API": {
            "method": "get",
            "endpoint": "/"
        },
        "rbac": {
            "conditions": {
                "operator": "$and",
                "criteria": [
                    {
                        "function": "Invalid",
                        "arguments": {
                            "field": "password",
                            "value": "password",
                            "custom": ""
                        }
                    }
                ]
            }
        }
    })), "condition", {
        "id": "12345",
        "username": "owner",
        "email": "mike@corsairm360.com",
        "account": "6041ed8e08acbf71381e0789",
        "primary": True,
        "config": {
            "test": "mike"
        }
    })
    assert not status

def test_can_access_conditions_fail_condition_parsing_error():
    status = rbac.can_access(build_request(to_json({
        "user": {
            "id": "12345",
            "username": "owner",
            "email": "mike@corsairm360.com",
            "account": {
                "_id": "6041ed8e08acbf71381e0789",
                "name": "mikeorg",
            }
        },
        "API": {
            "method": "get",
            "endpoint": "/"
        },
        "rbac": {
            "conditions": {
                "operator": "$and",
                "criteria": [
                    {
                        "function": "IN",
                        "arguments": {
                            "field": True,
                            "value": "test",
                            "custom": ""
                        }
                    }
                ]
            }
        }
    })), "conditions", {
        "id": "12345",
        "username": "owner",
        "email": "mike@corsairm360.com",
        "account": "6041ed8e08acbf71381e0789",
        "primary": True,
        "config": {
            "test": "mike"
        }
    })
    assert not status

def test_filter_fields_fail_invalid_data():
    data = {}
    status = rbac.filter_fields(build_request(to_json({
        "API": {
            "method": "get",
            "endpoint": "/"
        },
        "rbac": {
            "fields": {}
        }
    })), data)
    assert status == data

def test_filter_fields_success_no_rbac_configured():
    data = {"foo": "bar"}
    status = rbac.filter_fields(build_request(to_json({
        "API": {
            "method": "get",
            "endpoint": "/"
        },
        "rbac": {
            "fields": {
                "operator": None
            }
        }
    })), data)
    assert status == data

def test_filter_fields_success_will_work():
    data = {
        "_id": "1234",
        "username": "owner",
        "firstName": "owner",
        "lastName": "Maestro M360",
        "email": "mike@corsairm360.com",
        "status": "active",
        "profile": {},
        "ts": 1614693465902,
        "account": "45678",
        "groups": ["admin", "root"],
        "security": {"mfa": False},
        "tenants": [{"id": "abcd", "code": "M360"}],
        "config": {"M360": {"common": {"primary": True, "pet": "cersei", "foo": "bar"}, "specific": {}}},
        "auto_registration_flow": "corsair_saas"
    }
    status = rbac.filter_fields(build_request(to_json({
        "API": {
            "method": "get",
            "endpoint": "/"
        },
        "rbac": {
            "fields": {
                "operator": "allow",
                "list": {
                    "user": "username",
                    "email": "email",
                    "config": "config.M360.common"
                }
            }
        }
    })), data)
    assert status == {
        "_id": "1234",
        "username": "owner",
        "firstName": "owner",
        "lastName": "Maestro M360",
        "email": "mike@corsairm360.com",
        "status": "active",
        "profile": {},
        "ts": 1614693465902,
        "account": "45678",
        "groups": ["admin", "root"],
        "security": {"mfa": False},
        "tenants": [{"id": "abcd", "code": "M360"}],
        "config": {"M360": {"common": {"primary": True, "pet": "cersei", "foo": "bar"}}},
        "auto_registration_flow": "corsair_saas"
    }

def test_set_service_config_2():
    rbac.set_service_config({
        "contract": {
            "name": "express",
            "group": "Testing",
            "version": 1,
            "ports": {
                "data": 4002,
                "maintenance": 4002
            },
            "apis": {
                "main": {
                    "get": {
                        "/": {
                            "label": "get api",
                            "access": True
                        }
                    },
                    "post": {
                        "/": {
                            "access": False,
                            "label": "post api",
                            "rbac": {
                                "fields": ["config"]
                            }
                        }
                    }
                }
            }
        }
    })

def test_filter_fields_success_will_work_on_array_and_deny():
    data = [{
        "_id": "1234",
        "username": "owner",
        "firstName": "owner",
        "lastName": "Maestro M360",
        "email": "mike@corsairm360.com",
        "status": "active",
        "profile": {},
        "ts": 1614693465902,
        "account": "45678",
        "groups": ["admin", "root"],
        "security": {"mfa": False},
        "tenants": [{"id": "abcd", "code": "M360"}],
        "config": {"M360": {"common": {"primary": True, "pet": "cersei", "foo": "bar"}, "specific": {}}},
        "auto_registration_flow": "corsair_saas"
    }]
    status = rbac.filter_fields(build_request(to_json({
        "API": {
            "method": "get",
            "endpoint": "/"
        },
        "rbac": {
            "fields": {
                "operator": "deny",
                "list": {
                    "tenant": "tenants",
                    "security": "security",
                    "groups": "groups",
                    "config": "config.M360"
                }
            }
        }
    })), data)
    assert status == [
        {
            "_id": "1234",
            "username": "owner",
            "firstName": "owner",
            "lastName": "Maestro M360",
            "email": "mike@corsairm360.com",
            "status": "active",
            "profile": {},
            "ts": 1614693465902,
            "account": "45678",
            "groups": ["admin", "root"],
            "security": {"mfa": False},
            "tenants": [{"id": "abcd", "code": "M360"}],
            "config": {"M360": {"common": {"primary": True, "pet": "cersei", "foo": "bar"}, "specific": {}}},
            "auto_registration_flow": "corsair_saas"
        }
    ]



import pytest

from m360.registry import registry_instance, resource_instance, database_instance, custom_registry_instance

registry = registry_instance()
resource = resource_instance()
database = database_instance()
customreg = custom_registry_instance()

@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before each test
    registry.clear()  # clear cache so that individual tests don't overlap
    # A test function will be run at this point
    yield
    # Code that will run after each test

def test_get_registry(requests_mock):
    reg = {"name": "my_registry", "config": {"a": 1, "b": "2", "c": True}, "_TTL": 5000}
    # mock the gateway call
    requests_mock.get("https://127.0.0.1:5000/registry/", json={"result": True, "data": {"registry": reg}})

    res = registry.get()
    assert res == reg
    res = registry.get("config")
    assert res == reg.get("config")
    res = registry.get("_TTL")
    assert res == reg.get("_TTL")
    res = registry.get("invalid")  # should return the whole registry
    assert res == reg

def test_reload_registry(requests_mock):
    reg1 = {"name": "my_registry", "config": {"a": 1, "b": "2", "c": True}, "_TTL": 5000}
    # mock the gateway call
    requests_mock.get("https://127.0.0.1:5000/registry/", json={"result": True, "data": {"registry": reg1}})

    res = registry.get()
    assert res == reg1

    reg2 = {"name": "my_new_registry", "config": {"updated": True}, "_TTL": 6000}
    # mock the gateway call
    requests_mock.get("https://127.0.0.1:5000/registry/", json={"result": True, "data": {"registry": reg2}})

    res = registry.reload()
    assert res == reg2

def test_clear_registry(requests_mock):
    reg1 = {"name": "my_registry", "config": {"a": 1, "b": "2", "c": True}, "_TTL": 5000}
    # mock the gateway call
    requests_mock.get("https://127.0.0.1:5000/registry/", json={"result": True, "data": {"registry": reg1}})

    res = registry.reload()
    assert res == reg1

    # mock the gateway call
    reg2 = {"name": "my_new_registry", "config": {"updated": True}, "_TTL": 6000}
    requests_mock.get("https://127.0.0.1:5000/registry/", json={"result": True, "data": {"registry": reg2}})

    res = registry.get()
    assert res == reg1

    registry.clear()
    res = registry.get()
    assert res == reg2

def test_get_custom_registry(requests_mock):
    reg = {"name": "my_registry", "_TTL": 5000}
    cus = {"cust1": {"a": 1}, "cust2": {"b": "22"}, "cust3": {"c": {"d": "test"}}}
    # mock the gateway call
    requests_mock.get("https://127.0.0.1:5000/registry/", json={"registry": reg, "custom": cus})
    registry.load()

    res = customreg.get()
    assert res == cus
    res = customreg.get("cust1")
    assert res == cus.get("cust1")
    res = customreg.get("cust2")
    assert res == cus.get("cust2")
    res = customreg.get("cust3")
    assert res == cus.get("cust3")

def test_get_resources(requests_mock):
    reg = {"name": "my_registry", "_TTL": 5000}
    src = {"res1": {"a": 1}, "res2": {"b": "22"}, "res3": {"c": {"d": "test"}}}
    # mock the gateway call
    requests_mock.get("https://127.0.0.1:5000/registry/", json={"registry": reg, "resources": src})
    registry.load()

    res = resource.get()
    assert res == src
    res = resource.get("res1")
    assert res == src.get("res1")
    res = resource.get("res2")
    assert res == src.get("res2")
    res = resource.get("res3")
    assert res == src.get("res3")

def test_get_databases_single(requests_mock):
    reg = {"name": "my_registry", "_TTL": 5000}
    db = {"single": {"db1": {"name": "db1", "cluster": "cluster_11"}, "db2": {"name": "db2", "cluster": "cluster_22"}}}
    # mock the gateway call
    requests_mock.get("https://127.0.0.1:5000/registry/", json={"registry": reg, "databases": db})
    registry.load()
    db["single"]["db1"]["cluster"] = {"name": db["single"]["db1"]["cluster"]}
    db["single"]["db2"]["cluster"] = {"name": db["single"]["db2"]["cluster"]}

    res = database.get("single")
    assert res == db.get("single")
    res = database.get("single", "db1")
    assert res == db.get("single").get("db1")
    res = database.get("single", "db2", "tenant1")
    assert res == db.get("single").get("db2")

def test_get_databases_multi(requests_mock):
    db_name = "multi_db"
    reg = {"name": "my_registry", "_TTL": 5000}
    db = {"single": {"db1": {"name": "db1", "cluster": "cluster_11"}},
          "multitenant": {db_name: {"name": db_name, "cluster": "cluster_22"}}}
    # mock the gateway call
    requests_mock.get("https://127.0.0.1:5000/registry/", json={"registry": reg, "databases": db})
    registry.load()
    db["single"]["db1"]["cluster"] = {"name": db["single"]["db1"]["cluster"]}
    db["multitenant"][db_name]["cluster"] = {"name": db["multitenant"][db_name]["cluster"]}

    res = database.get("single")
    assert res == db.get("single")
    res = database.get("single", "db1", "tenant1")
    assert res == db.get("single").get("db1")
    tenant = "tenant1"
    db["multitenant"][db_name]["name"] = tenant + "_" + db_name
    res = database.get("multitenant", None, tenant)
    assert res == db.get("multitenant")
    res = database.get("multitenant", db_name, "tenant1")
    assert res == db.get("multitenant").get(db_name)



"""
Copyright (C) 2022 Corsair M360, Inc - All Rights Reserved.
Unauthorized copying of this file, via any medium is strictly prohibited.
Proprietary and confidential.
"""

import os
import json
import logging
import traceback
from pydash import py_

from m360 import utils, service, tenant, user
from m360.gateway import connector

_logger = logging.getLogger(__name__)
_connector = connector.instance()

class Rbac(object):
    """Offers rbac related helper methods"""

    # singleton
    _instance = None

    _config = None

    def __init__(self):
        _logger.debug("Rbac created")

    @staticmethod
    def rbac_singleton():
        """Access the singleton receiver"""

        if Rbac._instance:
            return Rbac._instance

        if not Rbac._instance:
            _logger.debug("Creating rbac instance")
            Rbac._instance = Rbac()

        return Rbac._instance

    @classmethod
    def set_service_config(cls, config):
        """
        This method sets the configuration of the middleware in the RBAC library
        @param config   {dict}
        """
        cls._config = config

    @classmethod
    def get(cls, request, section=None):
        """
        This method checks that m360 is provided in the request headers.
        Then it scans to find the fields' information inside m360 and returns it.
        @param request   {dict}
        @param section   {dict}
        @returns {dict}
        """
        if not request:
            return None
        headers = utils.get_req_headers(request)
        if not headers or not headers.get("m360"):
            return None
        m360 = headers.get("m360")
        try:
            if m360:
                m360 = json.loads(m360)
        except Exception as e:
            _logger.error(e)

        if not m360 or not utils.is_dict(m360) or not m360.get("rbac"):
            return None

        if section == "resource" or section == "resources":
            section = "resources"
        elif section == "condition" or section == "conditions":
            section = "conditions"
        elif section == "field" or section == "fields":
            section = "fields"
        elif section:
            raise Exception("Invalid section: " + section)

        if section and len(section.strip()) > 0 and m360.get("rbac").get(section):
            return m360.get("rbac").get(section)

        return m360.get("rbac")

    @classmethod
    def get_rbac_fields_from_service_contract(cls, request, rbac_field_options):
        """
        This method checks if the rbac fields configuration is set for a specific endpoint in the service contract
        by cross-referencing the request method and url with the contract
        @param request              {dict}
        @param rbac_field_options   {dict}
        @returns {dict}
        """
        if not rbac_field_options:
            return None

        rbac_field_options["allow"] = []
        rbac_field_options["deny"] = []
        endpoint_info = service.instance().get_endpoint_info(request)
        if not endpoint_info or not endpoint_info.get("method") or not endpoint_info.get("endpoint"):
            return rbac_field_options

        method = endpoint_info.get("method")
        endpoint = endpoint_info.get("endpoint")
        fields = py_.get(cls._config, 'contract.apis.main.' + method + '.' + endpoint + '.rbac.fields')
        if not fields:
            return rbac_field_options

        rbac_field_options["contract"] = fields
        rbac_configured_field_list = py_.clone(rbac_field_options.get("list"))
        processed_fields = []
        for contract_field in fields:
            if rbac_configured_field_list:
                for k, v in rbac_configured_field_list.items():
                    if v == contract_field or contract_field in v:
                        if rbac_field_options.get("operator") == "allow":
                            if v not in rbac_field_options["allow"]:
                                rbac_field_options["allow"].append(v)
                        else:
                            if v not in rbac_field_options["deny"]:
                                rbac_field_options["deny"].append(v)
                        processed_fields.append(contract_field)
        for contract_field in fields:
            if contract_field not in processed_fields:
                if rbac_field_options.get("operator") == "allow":
                    if contract_field not in rbac_field_options["deny"]:
                        rbac_field_options["deny"].append(contract_field)
                else:
                    if contract_field not in rbac_field_options["allow"]:
                        rbac_field_options["allow"].append(contract_field)

        return rbac_field_options

    @classmethod
    def can_access(cls, request, mode=None, data=None):
        """
        Simplified method that wraps around the other can Access operations
        @param request  {dict}
        @param mode     {str}
        @param data     {dict}
        @returns {bool}
        """
        options = None
        if mode in ["resource", "resources"]:
            options = cls.get(request, "resources")
            return cls.can_access_resources(request, options, data)
        elif mode in ["condition", "conditions"]:
            options = cls.get(request, "conditions")
            return cls.can_access_conditions(request, options, data)
        elif mode in ["field", "fields"]:
            options = cls.get(request, "fields")
            return cls.can_access_fields(request, options, data)
        else:
            raise Exception("Unsupported RBAC Access Mode Parameter 2!")

    @classmethod
    def can_access_fields(cls, request, options=None, data=None):
        """
        This method checks if the access is granted or denied for the given data object based on the given options
        @param request  {dict}
        @param options  {dict}
        @param data     {dict}
        @returns {bool}
        """
        # get the list of configured fields
        rbac_options = py_.clone_deep(options)

        def check_all_fields_in_object(data_obj):
            # loop in configured field and for each entry check in the object
            for i in data_obj:
                # ignore the non contract fields
                if i not in rbac_options.get("contract"):
                    continue
                # return false immediately if access is denied to any of the fields
                if i in rbac_options.get("deny"):
                    return False
            # return ok, all normal
            return True

        def is_field_rbac_configured():
            return rbac_options and rbac_options.get("operator") and rbac_options["operator"] in ["allow", "deny"] \
                   and rbac_options.get("list") and len(rbac_options["list"]) > 0

        if not data or not utils.is_dict(data):
            raise Exception("Unsupported data type provided, parameter 3 should can only be of type Object")

        # default to true if no RBAC fields is configured
        if not is_field_rbac_configured():
            return True

        rbac_options = cls.get_rbac_fields_from_service_contract(request, rbac_options)
        return check_all_fields_in_object(data)

    @classmethod
    def can_access_resources(cls, request, options=None, data=None):
        """
        This method checks if the user in the request has access to the data record provided based on Configured
        RBAC Resources
        @param request  {dict}
        @param options  {dict}
        @param data     {dict}
        @returns {bool}
        """

        def check_resources_access_in_object():
            # if no user, return false
            usr = user.instance().get(request)
            if not usr:
                return False
            # if mode is any, return true
            if options.get("mode") == "any":
                return True
            # get the value from the data record
            left_field = py_.get(data, options.get("field"))
            # get the value from the request user object
            right_field = py_.get(usr, options.get("value"))
            # return true if both match
            return left_field == right_field

        def is_field_rbac_configured():
            return options and options.get("mode") and options["mode"] in ["any", "own"] and \
                   (options["mode"] == "any" or (options["mode"] == "own" and options.get("field") and
                    len(options["field"].strip()) > 0 and options.get("value") and len(options["value"].strip()) > 0))

        if not data or utils.is_array(data) or not utils.is_dict(data):
            raise Exception("Unsupported data type provided, parameter 3 should can only be of type Object")

        # default to true if no RBAC fields is configured
        if not is_field_rbac_configured():
            return True

        return check_resources_access_in_object()

    @classmethod
    def can_access_conditions(cls, request, options=None, data=None):
        """
        This method checks if the user in the request has access to the data record provided based on Configured
        RBAC Conditions
        @param request  {dict}
        @param options  {dict}
        @param data     {dict}
        @returns {bool}
        """
        allowed_methods = ["EMPTY", "NOT_EMPTY", "EQ", "NOT_EQ", "START", "NOT_START", "END", "NOT_END", "IN",
                           "NOT_IN", "CONTAIN", "NOT_CONTAIN"]

        def check_one_condition(usr, condition):
            # get the value from the data record
            left_field = py_.get(data, condition["arguments"]["field"])

            if condition["arguments"].get("tCode") and len(condition["arguments"]["tCode"].strip()) > 0:
                # if no tenant, return false
                ten = tenant.instance().get(request, condition["arguments"]["tCode"])
                if not ten:
                    return False

                left_field = py_.get(ten, condition["arguments"]["field"])

            if condition["arguments"].get("custom") and len(condition["arguments"]["custom"].strip()) > 0:
                right_field = condition["arguments"]["custom"]
            else:
                # get the value from the request user object
                right_field = py_.get(usr, condition["arguments"]["value"])

            if right_field in ["true", "false"]:
                right_field = (right_field == "true")

            # return true if function matches
            condition_status_result = False
            cond_func = condition["function"].lower()
            if cond_func == "empty":
                if not left_field:
                    condition_status_result = True
            elif cond_func == "not_empty":
                if left_field:
                    condition_status_result = True
            elif cond_func == "eq":
                if left_field == right_field:
                    condition_status_result = True
            elif cond_func == "not_eq":
                if left_field != right_field:
                    condition_status_result = True
            elif cond_func == "start":
                if left_field.startswith(right_field):
                    condition_status_result = True
            elif cond_func == "not_start":
                if not left_field.startswith(right_field):
                    condition_status_result = True
            elif cond_func == "end":
                if left_field.endswith(right_field):
                    condition_status_result = True
            elif cond_func == "not_end":
                if not left_field.endswith(right_field):
                    condition_status_result = True
            elif cond_func in ["in", "contain"]:
                if utils.is_iterable(left_field) and right_field in left_field:
                    condition_status_result = True
            elif cond_func in ["not_in", "not_contain"]:
                if utils.is_iterable(left_field) and right_field not in left_field:
                    condition_status_result = True

            return condition_status_result

        def check_configured_conditions_for_access_in_object():
            # if no user, return false
            usr = user.instance().get(request)
            if not usr:
                return False

            conditions_met = []
            for criteria in options["criteria"]:
                if criteria.get("function") and criteria["function"] in allowed_methods and criteria.get("arguments") \
                        and "field" in criteria["arguments"] and "value" in criteria["arguments"]:
                    try:
                        conditions_met.append(check_one_condition(usr, criteria))
                    except Exception as e:
                        traceback.print_exc()
                        return e
                else:
                    # return false immediately if one condition doesn't have the correct format
                    return False

            # if operator is "and", all the values in the array must be true
            if options.get("operator") == "$and":
                return False not in conditions_met
            else:
                return True in conditions_met

        def is_field_rbac_configured():
            return options and options.get("operator") and options["operator"] in ["$and", "$or"] and \
                   options.get("criteria") and utils.is_array(options["criteria"]) and len(options["criteria"]) > 0

        if not data or utils.is_array(data) or not utils.is_dict(data):
            raise Exception("Unsupported data type provided, parameter 3 should can only be of type Object")

        if is_field_rbac_configured():
            return check_configured_conditions_for_access_in_object()
        else:
            return True

    @classmethod
    def filter_fields(cls, request, data=None):
        """
        This method removes the fields from the record if the access is false
        @param request  {dict}
        @param data     {dict or list}
        @returns {dict}
        """
        rbac_options = cls.get(request, "fields")
        options = cls.get_rbac_fields_from_service_contract(request, py_.clone_deep(rbac_options))

        def do_filter(record):
            return_record = {}
            if options.get("operator") == "deny":
                return_record = py_.clone(record)
                for denied in options.get("deny"):
                    utils.remove_object_from_path(return_record, denied)
            else:
                # fill the fields that are not in the contract
                for i in record:
                    if i not in options.get("contract"):
                        return_record[i] = py_.clone_deep(record[i])
                for allowed in options.get("allow"):
                    value = py_.get(record, allowed)
                    utils.build_object_from_path(return_record, allowed, py_.clone_deep(value))

            return return_record

        def is_field_rbac_configured():
            return options and options.get("operator") and options["operator"] in ["allow", "deny"] and \
                   options.get("list") and len(options["list"]) > 0

        if not is_field_rbac_configured():
            return data

        if data:
            if utils.is_array(data):
                for i in range(len(data)):
                    data[i] = do_filter(data[i])
            elif utils.is_dict(data):
                data = do_filter(data)

        return data


def instance():
    """Get the singleton instance of the rbac"""
    _logger.debug("calling instance, os.pid=%d", os.getpid())
    return Rbac.rbac_singleton()



"""
Copyright (C) 2022 Corsair M360, Inc - All Rights Reserved.
Unauthorized copying of this file, via any medium is strictly prohibited.
Proprietary and confidential.
"""

import json
import time
import logging
import requests

from m360 import constants

_logger = logging.getLogger(__name__)

def request(*, url, method="get", headers=None, query=None, body=None, is_json=False, verify=False):
    _logger.debug("Executing request: " + method + " " + url)
    if is_json:
        headers = headers if headers is not None else {}
        headers["Content-Type"] = "application/json"
    return requests.request(url=url, method=method, headers=headers, params=query, json=body, verify=verify)

def retry_request(*, url, method="get", headers=None, query=None, body=None, is_json=False, verify=False, retries=0,
                  retry_delay=constants.DEFAULT_RETRY_DELAY_SEC):
    counter = 0
    errors = ""
    ok = False
    main_res = None

    # make request with retry strategy
    while not ok:
        try:
            main_res = request(url=url, method=method, headers=headers, query=query, body=body, is_json=is_json,
                               verify=verify)
            res = main_res.json()
            # _logger.debug("Response: " + str(res))
            counter += 1
            # response is: { result: boolean, errors?: {}, data?: any }
            if res and isinstance(res, dict) and "result" in res:
                if res.get("result") is False or res.get("errors"):
                    if isinstance(res.get("errors"), dict):
                        for err in res.get("errors").values():
                            errors += err.get("message") + "\n"
                    if counter <= retries:
                        time.sleep(retry_delay)
                    else:
                        break
                else:
                    ok = True
                    break
            else:
                ok = True
                break
        except Exception as e:
            _logger.error(e)
            errors += str(e) + "\n"
            counter += 1
            if counter <= retries:
                time.sleep(retry_delay)
            else:
                break
    # return
    if ok:
        _logger.debug("Successfully called " + method.upper() + " " + url)
        return main_res
    else:
        _logger.error("Error(s) while calling " + method.upper() + " " + url)
        raise Exception(errors)

def json_file_to_dict(filepath):
    """
    Loads the file at the specified path and parses the JSON into a dict
    @param filepath  {str}  Path to a JSON file
    @returns {dict} The resulting dict
    """
    if not filepath or type(filepath) is not str or not filepath.strip():
        _logger.error("Invalid file path")
        raise Exception("Invalid file path")

    try:
        with open(filepath, mode="r", encoding="utf8") as f:
            data = f.read()
        data = json.loads(data)
        return data
    except Exception as e:
        _logger.error(e)
        raise e

def to_json(obj):
    return json.dumps(obj, separators=(",", ":"))

def from_json(text):
    return json.loads(text)

def get_req_headers(req):
    if hasattr(req, "headers"):
        return req.headers
    if hasattr(req, "META"):
        return req.META
    return {}

def is_array(obj):
    return isinstance(obj, list)

def is_dict(obj):
    return isinstance(obj, dict)

def is_string(obj):
    return isinstance(obj, str)

def is_iterable(obj):
    try:
        iterator = iter(obj)
    except TypeError:
        return False
    else:
        return True

def build_object_from_path(obj, path, value):
    paths = path.split(".")
    last = paths.pop()
    ref = obj
    for p in paths:
        if not ref.get(p):
            ref[p] = {}
        ref = ref[p]
    ref[last] = value

def remove_object_from_path(obj, path):
    paths = path.split(".")
    last = paths.pop()
    ref = obj
    for p in paths:
        if ref.get(p):
            ref = ref[p]
        else:
            return
    if ref.get(last):
        del ref[last]



import os
import time
import pytest

from m360.sdk import SDK, instance
from m360.awareness import Awareness
from m360.maintenance import Maintenance
from m360.service import Service
from m360.tenant import Tenant
from m360.user import User
from m360.rbac import Rbac
from m360.validator import Validator
from m360.registry import Database, Registry, Resource, CustomRegistry
from m360.frameworks.helper import Helper
from .helper import mock_django, Request

DIR_NAME = os.path.dirname(__file__)

@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before each test
    yield  # A test function will be run at this point
    # Code that will run after each test
    Helper._instance = None

def test_ssl_cert_error():
    with pytest.raises(SystemExit):
        mock_django()

def test_validation_errors():
    os.environ["APP_SSL_KEY"] = "true"
    os.environ["APP_SSL_CERT"] = "true"
    with pytest.raises(SystemExit):
        mock_django()
    with pytest.raises(SystemExit):
        mock_django({"invalid": "config"})
    with pytest.raises(SystemExit):
        mock_django({
            "contract": os.path.join(DIR_NAME, "../assets/contract.json"),
            "ip": "localhost",
            "type": "invalid",
            "platform": "manual"
        })
    with pytest.raises(SystemExit):
        mock_django({
            "contract": "invalid",
            "ip": "localhost",
            "type": "django",
            "platform": "manual"
        })
    with pytest.raises(SystemExit):
        mock_django({
            "contract": os.path.join(DIR_NAME, "../assets/contract.json"),
            "type": "django",
            "platform": "docker"
        })

def test_middleware(requests_mock):
    def handler(request):
        return request.GET
    # setup sdk
    os.environ["APP_SSL_KEY"] = "true"
    os.environ["APP_SSL_CERT"] = "true"
    reg = {"name": "my_registry", "_TTL": 10000}
    requests_mock.get("https://127.0.0.1:5000/registry/", json={"result": True, "data": {"registry": reg}})
    requests_mock.post("https://127.0.0.1:5000/service/register", json={"result": True, "data": {}})
    mw = mock_django(config={
        "contract": os.path.join(DIR_NAME, "../assets/contract.json"),
        "ip": "localhost",
        "type": "django",
        "platform": "manual"
    }, handler=handler)
    # wait for lazy init
    time.sleep(0.25)

    # verify SDK
    assert mw
    assert mw.helper
    assert isinstance(mw.helper.sdk, SDK)
    assert isinstance(mw.helper.maintenance, Maintenance)
    assert isinstance(mw.helper.sdk.tenant, Tenant)
    assert isinstance(mw.helper.sdk.user, User)
    assert isinstance(mw.helper.sdk.rbac, Rbac)
    assert isinstance(mw.helper.sdk.awareness, Awareness)
    assert isinstance(mw.helper.sdk.registry, Registry)
    assert isinstance(mw.helper.sdk.custom, CustomRegistry)
    assert isinstance(mw.helper.sdk.database, Database)
    assert isinstance(mw.helper.sdk.resource, Resource)
    assert isinstance(mw.helper.sdk.service, Service)
    assert isinstance(mw.helper.sdk.validator, Validator)
    # stop the timers (not needed for this test)
    mw.helper.sdk.stop_auto_reload()

    # test heartbeat
    req = Request(path="/heartbeat")
    res = mw(req)
    assert res
    assert res.get("heartbeat")
    assert res.get("name") == "mwtest"
    assert instance() is mw.helper.sdk

    # test awareness reload
    awareness = {"some-data": "goes here..."}
    requests_mock.get("https://127.0.0.1:5000/awareness", json={"result": True, "data": awareness})
    req = Request(path="/awareness/reload")
    res = mw(req)
    assert res
    assert res.get("awareness") == awareness
    assert res.get("name") == "mwtest"
    assert instance() is mw.helper.sdk

    # test registry reload
    req = Request(path="/registry/reload")
    res = mw(req)
    assert res
    assert res.get("registry") == reg
    assert res.get("name") == "mwtest"
    assert instance() is mw.helper.sdk

    # test regular routes
    req = Request(path="/not/special")
    params = {"q1": "string", "q2": 123, "q3": False}
    req.GET = params
    res = mw(req)
    assert res == params
    assert instance() is mw.helper.sdk



"""
Copyright (C) 2022 Corsair M360, Inc - All Rights Reserved.
Unauthorized copying of this file, via any medium is strictly prohibited.
Proprietary and confidential.
"""
import string
import random
import sys

from unittest import mock
from m360.utils import from_json

class Settings(object):
    def __init__(self, cfg=None):
        self.M360 = cfg

class DjangoConf(object):
    def __init__(self, cfg=None):
        self.settings = Settings(cfg)

class DjangoHttp(object):
    def __init__(self, resp):
        self.JsonResponse = resp

class Request(object):
    def __init__(self, env=None, *, headers=None, path=None, meta=None, method="get"):
        self.path = path
        self.method = method
        self.headers = headers
        self.META = meta
        self.M360 = None
        self.GET = None
        self.args = None
        if env:
            self.path = env.get("path") if env.get("path") else self.path
            self.method = env.get("method") if env.get("method") else self.method
            self.headers = env.get("headers") if env.get("headers") else self.headers
            self.args = env.get("args") if env.get("args") else self.args

class Response(object):
    def __init__(self, data, *, status=200, headers=None):
        data_res = from_json(data) if type(data) is str else data
        self.data = data_res
        self.status = status
        self.headers = headers

    def __call__(self, *args, **kwargs):
        # usual method call: res(environ, start_response)
        return args[1](self.data)

class WerkzeugWrappers(object):
    def __init__(self):
        self.Request = Request
        self.Response = Response

class Flask(object):
    def __init__(self):
        self.g = Settings()

class DockerContainer(object):
    def __init__(self, container):
        self.attrs = container

class DockerContainers(object):
    def __init__(self, containers):
        self.containers = []
        for c in containers:
            self.containers.append(DockerContainer(c))

    def list(self, filters=None):
        return self.containers

class Docker(object):
    def __init__(self, containers):
        self.containers = DockerContainers(containers)

    def from_env(self):
        return self

def random_string(length=10, uppercase=False):
    if uppercase:
        txt = string.ascii_uppercase
        return ''.join(random.choice(txt) for _ in range(length))
    else:
        txt = string.ascii_lowercase
        return ''.join(random.choice(txt) for _ in range(length))

def random_numeric_string(length=10):
    dig = string.digits
    return ''.join(random.choice(dig) for _ in range(length))

def random_int(_min=0, _max=10):
    return random.randint(_min, _max)

def mock_django(config=None, handler=None):
    def get_response(data=True, status=200):
        return data
    if not handler:
        handler = get_response
    try:
        sys.modules.pop("django")
        sys.modules.pop("django.http")
        sys.modules.pop("django.http.JsonResponse")
        sys.modules.pop("django.conf")
    except KeyError:
        pass  # ignore errors here
    sys.modules['django'] = mock.Mock()
    sys.modules['django.http'] = DjangoHttp(get_response)
    sys.modules['django.conf'] = DjangoConf(config)
    from m360.frameworks.django.middleware import DjangoMiddleware
    return DjangoMiddleware(handler)

def mock_flask(config=None, handler=None):
    def app(env, start_response):
        if handler:
            return start_response(handler(env))
        else:
            return start_response(env)
    try:
        sys.modules.pop("flask")
        sys.modules.pop("werkzeug")
        sys.modules.pop("werkzeug.wrappers")
    except KeyError:
        pass  # ignore errors here
    sys.modules['flask'] = Flask()
    sys.modules['werkzeug'] = mock.Mock()
    sys.modules['werkzeug.wrappers'] = WerkzeugWrappers()
    from m360.frameworks.flask.middleware import FlaskMiddleware
    return FlaskMiddleware(app, config)

def mock_docker(containers=None):
    try:
        sys.modules.pop("m360")
        sys.modules.pop("m360.maintenance")
        sys.modules.pop("docker")
    except KeyError:
        pass  # ignore errors here
    sys.modules["docker"] = Docker(containers)



import os
import sys
import pytest

from m360 import constants
from m360.registry import registry_instance
from m360.awareness import instance
from .helper import mock_docker

registry = registry_instance()
awareness = instance()
app_name = "my_app"
ns = "test_ns"
svc = "test_svc"
net = "test_net"
docker_version = "1.24"

constants.MW_TESTING = True

def get_maintenance(config):
    from m360.maintenance import Maintenance
    return Maintenance(config, app_name, registry, "my_service", awareness)

def test_reload_registry_manual(requests_mock):
    # init maintenance library
    maint = get_maintenance({"platform": "manual", "contract": {"ports": {"maintenance": 6000}}})
    reg = {"name": "my_registry", "_TTL": 5000}
    # mock the gateway call
    requests_mock.get("https://127.0.0.1:5000/registry/", json={"result": True, "data": {"registry": reg}})

    res = maint.reload_registry([])
    assert res
    assert res.get("name") == app_name
    assert res.get("registry") == reg

def test_reload_registry_kubernetes(requests_mock):
    # mock the kube API calls
    constants.SVC_ACCOUNT_TOKEN_PATH = os.path.join(os.path.dirname(__file__), "../assets/kube_token")
    requests_mock.get("https://kubernetes.default.svc/api/v1/namespaces/" + ns + "/endpoints/" + svc,
                      json={"subsets": [{"addresses": [{"ip": "10.0.0.1"}, {"ip": "10.0.0.2"}]},
                                        {"addresses": [{"ip": "10.0.0.3"}]}, {"addresses": [{}]}, {}]})
    # init maintenance library
    maint = get_maintenance({"platform": "kubernetes", "contract": {"ports": {"maintenance": 6000}},
                             "platformOptions": {"namespace": ns, "service": svc, "network": net}})
    reg = {"name": "my_registry", "_TTL": 5000}
    # mock the brother nodes and gateway calls
    requests_mock.get("https://10.0.0.1:6000/registry/reload", json={"result": True})
    requests_mock.get("https://10.0.0.2:6000/registry/reload", json={"result": True})
    requests_mock.get("https://10.0.0.3:6000/registry/reload", json={"result": True})
    requests_mock.get("https://127.0.0.1:5000/registry/", json={"result": True, "data": {"registry": reg}})

    res = maint.reload_registry([])
    assert res
    assert res.get("name") == app_name
    assert res.get("registry") == reg

def test_reload_registry_docker(requests_mock):
    # mock the docker library's API calls
    cont1 = {"Id": "1234", "NetworkSettings": {"Networks": {net: {"IPAddress": "10.0.0.1"}}}}
    cont2 = {"Id": "5678", "NetworkSettings": {"Networks": {net: {"IPAddress": "10.0.0.2"}}}}
    cont3 = {"Id": "91011", "NetworkSettings": {"Networks": {net: {"IPAddress": "10.0.0.3"}}}}
    containers = [cont1, cont2, cont3]
    mock_docker(containers)
    # init maintenance library
    maint = get_maintenance({"platform": "docker", "contract": {"ports": {"maintenance": 6000}},
                             "platformOptions": {"namespace": ns, "service": svc, "network": net}})
    reg = {"name": "my_registry", "_TTL": 5000}
    # mock the brother nodes and gateway calls
    requests_mock.get("https://10.0.0.1:6000/registry/reload", json={"result": True})
    requests_mock.get("https://10.0.0.2:6000/registry/reload", json={"result": True})
    requests_mock.get("https://10.0.0.3:6000/registry/reload", json={"result": True})
    requests_mock.get("https://127.0.0.1:5000/registry/", json={"result": True, "data": {"registry": reg}})

    res = maint.reload_registry([])
    assert res
    assert res.get("name") == app_name
    assert res.get("registry") == reg

def test_reload_awareness_1(requests_mock):
    # init maintenance library
    maint = get_maintenance({"platform": "manual", "contract": {"ports": {"maintenance": 6000}}})
    aware = {"name": "complete_awareness", "_TTL": 5000}
    # mock the gateway call
    requests_mock.get("https://127.0.0.1:5000/awareness", json={"result": True, "data": aware})

    res = maint.reload_awareness([])
    assert res
    assert res.get("name") == app_name
    assert res.get("awareness") == aware

def test_reload_awareness_kubernetes(requests_mock):
    # mock the kube API calls
    constants.SVC_ACCOUNT_TOKEN_PATH = os.path.join(os.path.dirname(__file__), "../assets/kube_token")
    requests_mock.get("https://kubernetes.default.svc/api/v1/namespaces/" + ns + "/endpoints/" + svc,
                      json={"subsets": [{"addresses": [{"ip": "10.0.0.1"}, {"ip": "10.0.0.2"}]},
                                        {"addresses": [{"ip": "10.0.0.3"}]}, {"addresses": [{}]}, {}]})
    # init maintenance library
    maint = get_maintenance({"platform": "kubernetes", "contract": {"ports": {"maintenance": 6000}},
                             "platformOptions": {"namespace": ns, "service": svc, "network": net}})
    aware = {"name": "complete_awareness", "_TTL": 5000}
    # mock the brother nodes and gateway calls
    requests_mock.get("https://10.0.0.1:6000/awareness/reload", json={"result": True})
    requests_mock.get("https://10.0.0.2:6000/awareness/reload", json={"result": True})
    requests_mock.get("https://10.0.0.3:6000/awareness/reload", json={"result": True})
    requests_mock.get("https://127.0.0.1:5000/awareness", json={"result": True, "data": aware})

    res = maint.reload_awareness([])
    assert res
    assert res.get("name") == app_name
    assert res.get("awareness") == aware

def test_reload_awareness_docker(requests_mock):
    # mock the docker library's API calls
    cont1 = {"Id": "1234", "NetworkSettings": {"Networks": {net: {"IPAddress": "10.0.0.1"}}}}
    cont2 = {"Id": "5678", "NetworkSettings": {"Networks": {net: {"IPAddress": "10.0.0.2"}}}}
    cont3 = {"Id": "91011", "NetworkSettings": {"Networks": {net: {"IPAddress": "10.0.0.3"}}}}
    containers = [cont1, cont2, cont3]
    requests_mock.get("http+unix:///var/run/docker.sock", json={"result": True})
    requests_mock.get("http+docker://localhost/version", json={"ApiVersion": docker_version})
    requests_mock.get("http+docker://localhost/v" + docker_version + "/containers/json", json=containers)
    requests_mock.get("http+docker://localhost/v" + docker_version + "/containers/" + cont1["Id"] + "/json", json=cont1)
    requests_mock.get("http+docker://localhost/v" + docker_version + "/containers/" + cont2["Id"] + "/json", json=cont2)
    requests_mock.get("http+docker://localhost/v" + docker_version + "/containers/" + cont3["Id"] + "/json", json=cont3)
    # init maintenance library
    maint = get_maintenance({"platform": "docker", "contract": {"ports": {"maintenance": 6000}},
                             "platformOptions": {"namespace": ns, "service": svc, "network": net}})
    aware = {"name": "complete_awareness", "_TTL": 5000}
    # mock the brother nodes and gateway calls
    requests_mock.get("https://10.0.0.1:6000/awareness/reload", json={"result": True})
    requests_mock.get("https://10.0.0.2:6000/awareness/reload", json={"result": True})
    requests_mock.get("https://10.0.0.3:6000/awareness/reload", json={"result": True})
    requests_mock.get("https://127.0.0.1:5000/awareness", json={"result": True, "data": aware})

    res = maint.reload_awareness([])
    assert res
    assert res.get("name") == app_name
    assert res.get("awareness") == aware

def test_gateway_error(requests_mock):
    # init maintenance library
    maint = get_maintenance({"platform": "manual", "contract": {"ports": {"maintenance": 6000}}})
    # mock the gateway call
    errors = {"520": {"message": "Gateway not available"}}
    requests_mock.get("https://127.0.0.1:5000/registry/", status_code=520, json={"result": False, "errors": errors})

    with pytest.raises(Exception) as e:
        maint.reload_registry([])
    assert "Gateway not available" in str(e)

def test_invalid_kube_token_error(requests_mock):
    # use the wrong path on purpose
    constants.SVC_ACCOUNT_TOKEN_PATH = "/invalid/token"
    # mock the kube API calls
    requests_mock.get("https://kubernetes.default.svc/api/v1/namespaces/" + ns + "/endpoints/" + svc,
                      json={"subsets": [{"addresses": [{"ip": "10.0.0.1"}, {"ip": "10.0.0.2"}]},
                                        {"addresses": [{"ip": "10.0.0.3"}]}, {"addresses": [{}]}, {}]})
    # init maintenance library
    maint = get_maintenance({"platform": "kubernetes", "contract": {"ports": {"maintenance": 6000}},
                             "platformOptions": {"namespace": ns, "service": svc, "network": net}})
    reg = {"name": "my_registry", "_TTL": 5000}
    # mock the gateway call
    requests_mock.get("https://127.0.0.1:5000/registry/", json={"result": True, "data": {"registry": reg}})

    with pytest.raises(Exception) as e:
        maint.reload_registry([])
    assert "No such file or directory" in str(e)

def test_kubernetes_api_error(requests_mock):
    # mock the kube API calls
    constants.SVC_ACCOUNT_TOKEN_PATH = os.path.join(os.path.dirname(__file__), "../assets/kube_token")
    requests_mock.get("https://kubernetes.default.svc/api/v1/namespaces/" + ns + "/endpoints/" + svc,
                      status_code=500, json="")
    # init maintenance library
    maint = get_maintenance({"platform": "kubernetes", "contract": {"ports": {"maintenance": 6000}},
                             "platformOptions": {"namespace": ns, "service": svc, "network": net}})
    aware = {"name": "complete_awareness", "_TTL": 5000}
    # mock the brother nodes and gateway calls
    requests_mock.get("https://10.0.0.1:6000/awareness/reload", json={"result": True})
    requests_mock.get("https://10.0.0.2:6000/awareness/reload", json={"result": True})
    requests_mock.get("https://10.0.0.3:6000/awareness/reload", json={"result": True})
    requests_mock.get("https://127.0.0.1:5000/awareness", json={"result": True, "data": aware})

    with pytest.raises(Exception) as e:
        maint.reload_awareness([])
    assert "Invalid Kubernetes API response" in str(e)

def test_config_error(requests_mock):
    # mock the docker library's API calls
    cont1 = {"Id": "1234", "NetworkSettings": {"Networks": {net: {"IPAddress": "10.0.0.1"}}}}
    cont2 = {"Id": "5678", "NetworkSettings": {"Networks": {net: {"IPAddress": "10.0.0.2"}}}}
    cont3 = {"Id": "91011", "NetworkSettings": {"Networks": {net: {"IPAddress": "10.0.0.3"}}}}
    containers = [cont1, cont2, cont3]
    requests_mock.get("http+unix:///var/run/docker.sock", json={"result": True})
    requests_mock.get("http+docker://localhost/version", json={"ApiVersion": docker_version})
    requests_mock.get("http+docker://localhost/v" + docker_version + "/containers/json", json=containers)
    requests_mock.get("http+docker://localhost/v" + docker_version + "/containers/" + cont1["Id"] + "/json", json=cont1)
    requests_mock.get("http+docker://localhost/v" + docker_version + "/containers/" + cont2["Id"] + "/json", json=cont2)
    requests_mock.get("http+docker://localhost/v" + docker_version + "/containers/" + cont3["Id"] + "/json", json=cont3)
    # init maintenance library
    maint = get_maintenance({"platform": "docker", "contract": {"ports": {"maintenance": 6000}}})
    aware = {"name": "complete_awareness", "_TTL": 5000}
    # mock the brother nodes and gateway calls
    requests_mock.get("https://10.0.0.1:6000/awareness/reload", json={"result": True})
    requests_mock.get("https://10.0.0.2:6000/awareness/reload", json={"result": True})
    requests_mock.get("https://10.0.0.3:6000/awareness/reload", json={"result": True})
    requests_mock.get("https://127.0.0.1:5000/awareness", json={"result": True, "data": aware})

    with pytest.raises(Exception) as e:
        maint.reload_awareness([])
    assert "Configuration must specify 'platformOptions'" in str(e)

def test_docker_env_error(requests_mock):
    try:
        sys.modules.pop("m360")
        sys.modules.pop("m360.maintenance")
        sys.modules.pop("docker")
    except KeyError:
        pass  # ignore errors here
    # init maintenance library
    maint = get_maintenance({"platform": "docker", "contract": {"ports": {"maintenance": 6000}},
                             "platformOptions": {"namespace": ns, "service": svc, "network": net}})
    reg = {"name": "my_registry", "_TTL": 5000}
    # mock the gateway call
    requests_mock.get("https://127.0.0.1:5000/registry/", json={"result": True, "data": {"registry": reg}})

    with pytest.raises(Exception) as e:
        maint.reload_registry([])
    assert "Error while fetching server API version" in str(e)



"""
Copyright (C) 2022 Corsair M360, Inc - All Rights Reserved.
Unauthorized copying of this file, via any medium is strictly prohibited.
Proprietary and confidential.
"""

import os
import json
import logging

from m360 import utils
from m360.gateway import connector

_logger = logging.getLogger(__name__)
_connector = connector.instance()

class User(object):
    """Offers user related helper methods"""

    # singleton
    _instance = None

    def __init__(self):
        _logger.debug("User created")

    @staticmethod
    def user_singleton():
        """Access the singleton receiver"""

        if User._instance:
            return User._instance

        if not User._instance:
            _logger.debug("Creating user instance")
            User._instance = User()

        return User._instance

    @staticmethod
    def get(request):
        if not request:
            return None
        headers = utils.get_req_headers(request)
        if not headers or not headers.get("m360"):
            return None
        try:
            m360 = headers.get("m360")
            if m360:
                m360 = json.loads(m360)
        except Exception as e:
            _logger.error(e)
            m360 = None

        return m360.get("user") if m360 and type(m360) is dict and m360.get("user") else None

    @staticmethod
    def find(context):
        user_id = str(context.get("id")) if context.get("id") else None
        start = context.get("start") if context.get("start") else 1
        limit = context.get("limit") if context.get("limit") else 100
        group = context.get("group")
        service = context.get("service")
        version = context.get("version")
        tenant = context.get("tenant")
        url = "/users/" + user_id if user_id else "/users"
        qs = {}

        if not user_id:
            qs["start"] = start
            qs["limit"] = limit
        if group:
            qs["group"] = group
        if tenant:
            qs["tenant"] = tenant
        if service:
            qs["service"] = service
        if version:
            qs["version"] = version

        return _connector.invoke({"route": url, "qs": qs})

def instance():
    """Get the singleton instance of the user"""
    _logger.debug("calling instance, os.pid=%d", os.getpid())
    return User.user_singleton()



import os
import pytest

from m360 import constants
from m360.service import instance

constants.MW_TESTING = True

service = instance()
DIR_NAME = os.path.dirname(__file__)

@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before each test
    yield  # A test function will be run at this point
    # Code that will run after each test

def get_config(contract, ip, mwtype, platform, platform_options=None):
    cfg = {
        "ip": ip,
        "type": mwtype,
        "platform": platform
    }
    if contract:
        if type(contract) is str:
            cfg["contract"] = os.path.join(DIR_NAME, "../assets/" + contract)
        else:
            cfg["contract"] = contract
    if platform_options:
        cfg["platformOptions"] = platform_options
    return cfg

def test_register_django_manual(requests_mock):
    config = get_config("contract.json", "localhost", "django", "manual")
    requests_mock.post("https://127.0.0.1:5000/service/register", json={"result": True, "data": config})
    res = service.register(config.get("contract"), config.get("ip"), config)
    assert res == config

def test_register_django_docker(requests_mock):
    config = get_config("contract.json", "localhost", "django", "docker")
    requests_mock.post("https://127.0.0.1:5000/service/register", json={"result": True, "data": config})
    res = service.register(config.get("contract"), config.get("ip"), config)
    assert res == config

def test_register_django_kube(requests_mock):
    config = get_config("contract.json", "localhost", "django", "kubernetes", {"exposedPort": 7000})
    requests_mock.post("https://127.0.0.1:5000/service/register", json={"result": True, "data": config})
    res = service.register(config.get("contract"), config.get("ip"), config)
    assert res == config

def test_register_errors():
    config = get_config(None, "127.0.0.1", "django", "manual")
    with pytest.raises(Exception) as e:
        service.register(config.get("contract"), config.get("ip"), config)
    assert "Missing Service Contract!" in str(e)

    config = get_config(" ", "127.0.0.1", "django", "manual")
    with pytest.raises(Exception) as e:
        service.register(config.get("contract"), config.get("ip"), config)
    assert "No such file or directory" in str(e)

    config = get_config("invalid", "127.0.0.1", "django", "manual")
    with pytest.raises(Exception) as e:
        service.register(config.get("contract"), config.get("ip"), config)
    assert "No such file or directory" in str(e)

    contract = {"test": 1, "module": os}
    config = get_config(contract, "127.0.0.1", "django", "manual")
    with pytest.raises(Exception) as e:
        service.register(config.get("contract"), config.get("ip"), config)
    assert "Unable to parse the service contract that you provided" in str(e)

    config = get_config("contract.json", None, "django", "manual")
    with pytest.raises(Exception) as e:
        service.register(config.get("contract"), config.get("ip"), config)
    assert "Missing IP Address Value!" in str(e)

    config = get_config("contract.json", "  ", "django", "manual")
    with pytest.raises(Exception) as e:
        service.register(config.get("contract"), config.get("ip"), config)
    assert "Missing IP Address Value!" in str(e)

    config = get_config("contract.json", "localhost", "django", "kubernetes")
    with pytest.raises(Exception) as e:
        service.register(config.get("contract"), config.get("ip"), config)
    assert "Missing 'platformOptions' in config" in str(e)

    config = get_config("contract.json", "localhost", "django", "manual")
    with pytest.raises(Exception) as e:
        service.register(config.get("contract"), config.get("ip"), config)
    assert "Connection refused" in str(e)

def test_get_str(requests_mock):
    config = get_config("contract.json", "localhost", "django", "manual")
    requests_mock.get("https://127.0.0.1:5000/service", json={"result": True, "data": config})
    res = service.get(config.get("contract"))
    assert res == config

def test_get_dict(requests_mock):
    config = get_config({"name": "mw-test", "group": "M360GRP"}, "localhost", "django", "manual")
    requests_mock.get("https://127.0.0.1:5000/service", json={"result": True, "data": config})
    res = service.get(config.get("contract"))
    assert res == config

def test_get_errors():
    with pytest.raises(Exception) as e:
        service.get(None)
    assert "Missing Service Contract!" in str(e)

    with pytest.raises(Exception) as e:
        service.get("  ")
    assert "Missing Service Contract!" in str(e)

    with pytest.raises(Exception) as e:
        service.get("/invalid")
    assert "No such file or directory" in str(e)

    contract = {"test": 1, "module": os}
    with pytest.raises(Exception) as e:
        service.get(contract)
    assert "Unable to parse the service contract that you provided" in str(e)

    config = get_config("contract.json", "localhost", "django", "manual")
    with pytest.raises(Exception) as e:
        service.get(config.get("contract"))
    assert "Connection refused" in str(e)

def test_ensure_ip_address():
    res = service.ensure_ip_address({"platform": "manual", "ip": "http://www.google.com"})
    assert res
    res = service.ensure_ip_address({"platform": "kubernetes"})
    assert res

    with pytest.raises(Exception) as e:
        service.ensure_ip_address(None)
    assert "Invalid configuration" in str(e)
    with pytest.raises(Exception) as e:
        service.ensure_ip_address("")
    assert "Invalid configuration" in str(e)



"""
Copyright (C) 2022 Corsair M360, Inc - All Rights Reserved.
Unauthorized copying of this file, via any medium is strictly prohibited.
Proprietary and confidential.
"""

import os
import re
import logging

from m360 import constants, utils

_logger = logging.getLogger(__name__)
_ip_pattern = re.compile("^([0-9]{1,3}\\.){3}[0-9]{1,3}$")

class Connector(object):
    """Handles communication with the gateway"""

    # singleton
    _instance = None

    _service_name = None
    _service_version = None

    def __init__(self):
        _logger.debug("Connector created")

    def init(self, service_name, service_version):
        self._service_name = service_name
        self._service_version = service_version
        _logger.debug("Connector init complete")

    @staticmethod
    def connector_singleton():
        """Access the singleton receiver"""

        if Connector._instance:
            return Connector._instance

        if not Connector._instance:
            _logger.debug("Creating connector instance")
            Connector._instance = Connector()

        return Connector._instance

    def invoke(self, context=None):
        """Invokes gateway APIs"""
        _logger.debug("Invoking connector with the context: " + str(context))

        if not context:
            context = {}
        ip = str(context.get("ip")) if context.get("ip") else constants.GATEWAY_MAINTENANCE_IP
        ip = ip.replace("https://", "").replace("http://", "").replace("/", "")
        port = context.get("port") if context.get("port") else constants.GATEWAY_MAINTENANCE_PORT
        http_method = context.get("method").lower() if context.get("method") else "get"
        route = context.get("route") if context.get("route") else ""

        # build uri
        if ip != "localhost" and not _ip_pattern.match(ip):
            url = "https://" + ip + route
        else:
            url = "https://" + ip + ":" + str(port) + route

        # build headers
        headers = context.get("headers") if context.get("headers") else {}
        if self._service_name:
            headers["service"] = self._service_name
        if self._service_version:
            headers["version"] = str(self._service_version)

        query = context.get("qs") if context.get("qs") else None
        body = context.get("body") if context.get("body") else None

        max_count = int(context.get("retryOnFailure")) if context.get("retryOnFailure") else 1
        delay = int(context.get("retryDelay")) if context.get("retryDelay") else 5

        # make request with retry strategy
        res = utils.retry_request(url=url, method=http_method, headers=headers, query=query, body=body, is_json=True,
                                  retries=(max_count - 1), retry_delay=delay).json()
        if "result" in res and "data" in res:
            return res.get("data")
        else:
            return res

def instance():
    """Get the singleton instance of the connector"""
    _logger.debug("calling instance, os.pid=%d", os.getpid())
    return Connector.connector_singleton()
